# -*- coding: utf-8 -*-

if __name__ == "__main__":
    raise Exception(
        "Test files can't be run directly. Use `python -m pytest greenery`"
    )

from .bound import Bound, INF
from .charclass import Charclass, DOT, NULLCHARCLASS, DIGIT
from .multiplier import Multiplier, ONE, STAR, PLUS
from .rxelems import Mult, Conc, Pattern
from .parse import NoMatch, match_charclass, parse, match_mult


def test_charclass_matching():
    assert match_charclass("a", 0) == (Charclass("a"), 1)
    assert match_charclass("aa", 1) == (Charclass("a"), 2)
    assert match_charclass("a$", 1) == (Charclass("$"), 2)
    assert match_charclass(".", 0) == (DOT, 1)
    try:
        match_charclass("[", 0)
        assert False
    except IndexError:
        pass
    try:
        match_charclass("a", 1)
        assert False
    except NoMatch:
        pass
    assert match_charclass("[\\d]", 0) == (DIGIT, 4)


def test_negatives_inside_charclasses():
    assert match_charclass("[\\D]", 0) == (~DIGIT, 4)
    assert match_charclass("[a\\D]", 0) == (~DIGIT, 5)
    assert match_charclass("[a1\\D]", 0) == (~Charclass("023456789"), 6)
    assert match_charclass("[1a\\D]", 0) == (~Charclass("023456789"), 6)
    assert match_charclass("[1\\D]", 0) == (~Charclass("023456789"), 5)
    assert match_charclass("[\\Da]", 0) == (~DIGIT, 5)
    assert match_charclass("[\\D1]", 0) == (~Charclass("023456789"), 5)
    assert match_charclass("[\\D1a]", 0) == (~Charclass("023456789"), 6)
    assert match_charclass("[\\D\\d]", 0) == (DOT, 6)
    assert match_charclass("[\\D\\D]", 0) == (~DIGIT, 6)
    assert match_charclass("[\\S\\D]", 0) == \
        (~Charclass("\t\n\v\f\r 0123456789"), 6)
    assert match_charclass("[\\S \\D]", 0) == \
        (~Charclass("\t\n\v\f\r0123456789"), 7)


def test_negated_negatives_inside_charclasses():
    assert match_charclass("[^\\D]", 0) == (DIGIT, 5)
    assert match_charclass("[^a\\D]", 0) == (DIGIT, 6)
    assert match_charclass("[^a1\\D]", 0) == (Charclass("023456789"), 7)
    assert match_charclass("[^1a\\D]", 0) == (Charclass("023456789"), 7)
    assert match_charclass("[^1\\D]", 0) == (Charclass("023456789"), 6)
    assert match_charclass("[^\\Da]", 0) == (DIGIT, 6)
    assert match_charclass("[^\\D1]", 0) == (Charclass("023456789"), 6)
    assert match_charclass("[^\\D1a]", 0) == (Charclass("023456789"), 7)
    assert match_charclass("[^\\D\\d]", 0) == (NULLCHARCLASS, 7)
    assert match_charclass("[^\\D\\D]", 0) == (DIGIT, 7)
    assert match_charclass("[^\\S\\D]", 0) == \
        (Charclass("\t\n\v\f\r 0123456789"), 7)
    assert match_charclass("[^\\S \\D]", 0) == \
        (Charclass("\t\n\v\f\r0123456789"), 8)


def test_mult_matching():
    assert match_mult("abcde[^fg]*", 5) == (
        Mult(~Charclass("fg"), STAR),
        11
    )
    assert match_mult("abcde[^fg]*h{5}[a-z]+", 11) == (
        Mult(Charclass("h"), Multiplier(Bound(5), Bound(5))),
        15
    )
    assert match_mult("abcde[^fg]*h{5}[a-z]+T{1,}", 15) == (
        Mult(Charclass("abcdefghijklmnopqrstuvwxyz"), PLUS),
        21
    )
    assert match_mult("abcde[^fg]*h{5}[a-z]+T{2,}", 21) == (
        Mult(Charclass("T"), Multiplier(Bound(2), INF)),
        26
    )


def test_charclass_ranges():
    # Should accept arbitrary ranges of characters in charclasses. No longer
    # limited to alphanumerics. (User beware...)
    assert parse("[z{|}~]") == parse("[z-~]")
    assert parse("[\\w:;<=>?@\\[\\\\\\]\\^`]") == parse("[0-z]")


def test_hex_escapes():
    # Should be able to parse e.g. "\\x40"
    assert parse("\\x00") == parse("\x00")
    assert parse("\\x40") == parse("@")
    assert parse("[\\x40]") == parse("[@]")
    assert parse("[\\x41-\\x5a]") == parse("[A-Z]")


def test_w_d_s():
    # Allow "\w", "\d" and "\s" in charclasses
    assert parse("\\w") == parse("[0-9A-Z_a-z]")
    assert parse("[\\w~]") == parse("[0-9A-Z_a-z~]")
    assert parse("[\\da]") == parse("[0123456789a]")
    assert parse("[\\s]") == parse("[\t\n\r\f\v ]")


def test_mult_parsing():
    assert parse("[a-g]+") == Pattern(
        Conc(
            Mult(
                Charclass("abcdefg"),
                PLUS
            )
        )
    )
    assert parse("[a-g0-8$%]+") == Pattern(
        Conc(
            Mult(
                Charclass("abcdefg012345678$%"),
                PLUS
            )
        )
    )
    assert parse("[a-g0-8$%\\^]+") == Pattern(
        Conc(
            Mult(
                Charclass("abcdefg012345678$%^"),
                PLUS
            )
        )
    )


def test_conc_parsing():
    assert parse("abcde[^fg]*h{5}[a-z]+") == Pattern(
        Conc(
            Mult(Charclass("a"), ONE),
            Mult(Charclass("b"), ONE),
            Mult(Charclass("c"), ONE),
            Mult(Charclass("d"), ONE),
            Mult(Charclass("e"), ONE),
            Mult(~Charclass("fg"), STAR),
            Mult(Charclass("h"), Multiplier(Bound(5), Bound(5))),
            Mult(Charclass("abcdefghijklmnopqrstuvwxyz"), PLUS),
        )
    )
    assert parse("[bc]*[ab]*") == Pattern(
        Conc(
            Mult(Charclass("bc"), STAR),
            Mult(Charclass("ab"), STAR),
        )
    )
    assert parse("abc...") == Pattern(
        Conc(
            Mult(Charclass("a"), ONE),
            Mult(Charclass("b"), ONE),
            Mult(Charclass("c"), ONE),
            Mult(DOT, ONE),
            Mult(DOT, ONE),
            Mult(DOT, ONE),
        )
    )
    assert parse("\\d{4}-\\d{2}-\\d{2}") == Pattern(
        Conc(
            Mult(DIGIT, Multiplier(Bound(4), Bound(4))),
            Mult(Charclass("-"), ONE),
            Mult(DIGIT, Multiplier(Bound(2), Bound(2))),
            Mult(Charclass("-"), ONE),
            Mult(DIGIT, Multiplier(Bound(2), Bound(2))),
        )
    )


def test_pattern_parsing():
    assert parse("abc|def(ghi|jkl)") == Pattern(
        Conc(
            Mult(Charclass("a"), ONE),
            Mult(Charclass("b"), ONE),
            Mult(Charclass("c"), ONE),
        ),
        Conc(
            Mult(Charclass("d"), ONE),
            Mult(Charclass("e"), ONE),
            Mult(Charclass("f"), ONE),
            Mult(
                Pattern(
                    Conc(
                        Mult(Charclass("g"), ONE),
                        Mult(Charclass("h"), ONE),
                        Mult(Charclass("i"), ONE),
                    ),
                    Conc(
                        Mult(Charclass("j"), ONE),
                        Mult(Charclass("k"), ONE),
                        Mult(Charclass("l"), ONE),
                    ),
                ),
                ONE,
            ),
        ),
    )

    # Accept the "non-capturing group" syntax, "(?: ... )" but give it no
    # special significance
    assert parse("(?:)") == parse("()")
    assert parse("(?:abc|def)") == parse("(abc|def)")
    parse("(:abc)")  # should give no problems

    # Named groups
    assert parse("(?P<ng1>abc)") == parse("(abc)")



# -*- coding: utf-8 -*-

from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class Bound:
    '''An integer but sometimes also possibly infinite (None)'''
    v: Optional[int]

    def __post_init__(self):
        if self.v is not None and self.v < 0:
            raise Exception(f"Invalid bound: {repr(self.v)}")

    def __repr__(self):
        return f"Bound({repr(self.v)})"

    def __str__(self):
        if self == INF:
            # This only happens for an unlimited upper bound
            return ""
        return str(self.v)

    def __eq__(self, other):
        return self.v == other.v

    def __hash__(self):
        return hash(self.v)

    def __lt__(self, other):
        if self == INF:
            return False
        if other == INF:
            return True
        return self.v < other.v

    def __ge__(self, other):
        return not self < other

    def __mul__(self, other):
        '''Multiply this bound by another'''
        if self == Bound(0) or other == Bound(0):
            return Bound(0)
        if self == INF or other == INF:
            return INF
        return Bound(self.v * other.v)

    def __add__(self, other):
        '''Add this bound to another'''
        if self == INF or other == INF:
            return INF
        return Bound(self.v + other.v)

    def __sub__(self, other):
        '''
            Subtract another bound from this one.
            Caution: this operation is not meaningful for all bounds.
        '''
        if other == INF:
            if self != INF:
                raise Exception(
                    f"Can't subtract {repr(other)} from {repr(self)}"
                )

            # Infinity minus infinity is zero. This has to be true so that
            # we can for example subtract Multiplier(Bound(0), INF) from
            # Multiplier(Bound(1), INF) to get Multiplier(Bound(1), Bound(1))
            return Bound(0)
        if self == INF:
            return self
        return Bound(self.v - other.v)

    def copy(self):
        return Bound(self.v)


# Use this for cases where no upper bound is needed
INF = Bound(None)



# -*- coding: utf-8 -*-

'''
    Because of the circularity between `Pattern`, `Conc` and `Mult`, all three
    need to be in the same source file?
'''

from dataclasses import dataclass
from typing import Union

from .fsm import Fsm, ANYTHING_ELSE, null, epsilon, alphabet_key
from .multiplier import Multiplier, ZERO, QM, ONE, STAR
from .charclass import Charclass, NULLCHARCLASS
from .bound import Bound, INF


@dataclass(frozen=True)
class Conc():
    '''
        A `Conc` (short for "concatenation") is a tuple of `Mult`s i.e. an
        unbroken string of mults occurring one after the other.
        e.g. abcde[^fg]*h{4}[a-z]+(subpattern)(subpattern2)
        To express the empty string, use an empty `Conc`, Conc().
    '''
    def __init__(self, *mults):
        object.__setattr__(self, "mults", tuple(mults))

    def __eq__(self, other):
        return self.mults == other.mults

    def __hash__(self):
        return hash(self.mults)

    def __repr__(self):
        args = ", ".join(repr(mult) for mult in self.mults)
        return f"Conc({args})"

    def reduce(self):
        if self == NULLCONC:
            return self

        if self.empty():
            return NULLCONC

        # Try recursively reducing our mults
        reduced = tuple(mult.reduce() for mult in self.mults)
        if reduced != self.mults:
            return Conc(*reduced).reduce()

        # strip out mults which can only match the empty string
        for i, mult in enumerate(self.mults):
            if (
                # Conc contains "()" (i.e. a `Mult` containing only a `Pattern`
                # containing the empty string)? That can be removed
                # e.g. "a()b" -> "ab"
                (
                    mult.multiplicand == Pattern(EMPTYSTRING)
                ) \

                # If a `Mult` has an empty multiplicand, we can only match it
                # zero times => empty string => remove it entirely
                # e.g. "a[]{0,3}b" -> "ab"
                or (
                    mult.multiplicand.empty()
                    and mult.multiplier.min == Bound(0)
                ) \

                # Failing that, we have a positive multiplicand which we
                # intend to match zero times. In this case the only possible
                # match is the empty string => remove it
                # e.g. "a[XYZ]{0}b" -> "ab"
                or mult.multiplier == ZERO
            ):
                new = self.mults[:i] + self.mults[i + 1:]
                return Conc(*new).reduce()

        # We might be able to combine some mults together or at least simplify
        # the multiplier on one of them.
        if len(self.mults) > 1:
            for i in range(len(self.mults) - 1):
                r = self.mults[i]
                s = self.mults[i + 1]

                def to_pattern(multiplicand):
                    if isinstance(multiplicand, Pattern):
                        return multiplicand
                    return Pattern(Conc(Mult(multiplicand, ONE)))

                # so we can do intersection
                rm_pattern = to_pattern(r.multiplicand)
                sm_pattern = to_pattern(s.multiplicand)
                rm_sm_intersection = None

                # If R = S, then we can squish the multipliers together
                # e.g. ab?b?c -> ab{0,2}c
                if rm_pattern == sm_pattern:
                    squished = Mult(
                        rm_pattern,
                        r.multiplier + s.multiplier
                    )
                    new = self.mults[:i] + (squished,) + self.mults[i + 2:]
                    return Conc(*new).reduce()

                # If R's language is a subset of S's, then R{a,b}S{c,} reduces
                # to R{a}S{c,}...
                # e.g. \d+\w+ -> \d\w+
                # Do the cheapest checks first
                if r.multiplier.min < r.multiplier.max \
                   and s.multiplier.max == INF:
                    rm_sm_intersection = rm_pattern & sm_pattern
                    if rm_sm_intersection.equivalent(rm_pattern):
                        trimmed = Mult(
                            rm_pattern,
                            Multiplier(r.multiplier.min, r.multiplier.min)
                        )
                        new = self.mults[:i] + \
                            (trimmed, s) + \
                            self.mults[i + 2:]
                        return Conc(*new).reduce()

                # Conversely, if R is superset of S, then R{c,}S{a,b} reduces
                # to R{c,}S{a}.
                # e.g. [ab]+a? -> [ab]+
                # Do the cheapest checks first
                if r.multiplier.max == INF \
                   and s.multiplier.min < s.multiplier.max:
                    if rm_sm_intersection is None:
                        rm_sm_intersection = rm_pattern & sm_pattern
                    if rm_sm_intersection.equivalent(sm_pattern):
                        trimmed = Mult(
                            sm_pattern,
                            Multiplier(s.multiplier.min, s.multiplier.min)
                        )
                        new = self.mults[:i] + \
                            (r, trimmed) + \
                            self.mults[i + 2:]
                        return Conc(*new).reduce()

        # Conc contains (among other things) a *singleton* `Mult` containing
        # `Pattern` with only one internal `Conc`? Flatten out.
        # e.g. "a(d(ab|a*c))" -> "ad(ab|a*c)"
        # BUT NOT "a(d(ab|a*c)){2,}"
        # AND NOT "a(d(ab|a*c)|y)"
        for i, mult in enumerate(self.mults):
            if mult.multiplier == ONE \
               and isinstance(mult.multiplicand, Pattern) \
               and len(mult.multiplicand.concs) == 1:
                (conc,) = mult.multiplicand.concs
                new = self.mults[:i] + conc.mults + self.mults[i + 1:]
                return Conc(*new).reduce()

        return self

    def to_fsm(self, alphabet=None):
        if alphabet is None:
            alphabet = self.alphabet()

        # start with a component accepting only the empty string
        fsm1 = epsilon(alphabet)
        for mult in self.mults:
            fsm1 += mult.to_fsm(alphabet)
        return fsm1

    def alphabet(self):
        return {ANYTHING_ELSE}.union(*[mult.alphabet() for mult in self.mults])

    def empty(self):
        return any(mult.empty() for mult in self.mults)

    def __str__(self):
        return "".join(str(m) for m in self.mults)

    def common(self, other, suffix=False):
        '''
            Return the common prefix of these two `Conc`s; that is, the largest
            `Conc` which can be safely beheaded() from the front of both. The
            result could be `EMPTYSTRING`.
            "ZYAA, ZYBB" -> "ZY"
            "CZ, CZ" -> "CZ"
            "YC, ZC" -> ""

            With the "suffix" flag set, works from the end. E.g.:
            "AAZY, BBZY" -> "ZY"
            "CZ, CZ" -> "CZ"
            "CY, CZ" -> ""
        '''
        mults = []

        indices = range(min(len(self.mults), len(other.mults)))
        # e.g. [0, 1, 2, 3]

        # Work backwards from the end of both `Conc`s instead.
        if suffix:
            indices = [-i - 1 for i in indices]  # e.g. [-1, -2, -3, -4]

        for i in indices:
            x = self.mults[i]
            y = other.mults[i]
            common = x.common(y)

            # Happens when multiplicands disagree (e.g. "A.common(B)") or if
            # the multiplicand is shared but the common multiplier is `ZERO`
            # (e.g. "ABZ*.common(CZ)".)
            if common.multiplier == ZERO:
                break

            mults.append(common)

            # If we did not remove the entirety of both mults, we cannot
            # continue.
            if common != x or common != y:
                break

        if suffix:
            mults = reversed(mults)

        return Conc(*mults)

    def dock(self, other):
        '''
            Subtract another `Conc` from this one.
            This is the opposite of concatenation.
            For example, if ABC + DEF = ABCDEF,
            then logically ABCDEF - DEF = ABC.
        '''

        # e.g. self has mults at indices [0, 1, 2, 3, 4, 5, 6] len=7
        # e.g. other has mults at indices [0, 1, 2] len=3
        new = list(self.mults)
        for i in reversed(range(len(other.mults))):  # [2, 1, 0]
            # e.g. i = 1, j = 7 - 3 + 1 = 5
            j = len(self.mults) - len(other.mults) + i
            new[j] = new[j].dock(other.mults[i])

            if new[j].multiplier == ZERO:
                # omit that `Mult` entirely since it has been factored out
                del new[j]

            # If the subtraction is incomplete but there is more to
            # other.mults, then we have a problem. For example, "ABC{2} - BC"
            # subtracts the C successfully but leaves something behind,
            # then tries to subtract the B too, which isn't possible
            else:
                if i != 0:
                    raise Exception(
                        f"Can't subtract {repr(other)} from {repr(self)}"
                    )

        return Conc(*new)

    def behead(self, other):
        '''
            As with dock() but the other way around. For example, if
            ABC + DEF = ABCDEF, then ABCDEF.behead(AB) = CDEF.
        '''
        # Observe that FEDCBA - BA = FEDC.
        return self.reversed().dock(other.reversed()).reversed()

    def reversed(self):
        return Conc(*reversed([mult.reversed() for mult in self.mults]))


def from_fsm(f: Fsm):
    '''
        Turn the supplied finite state machine into a `Pattern`. This is
        accomplished using the Brzozowski algebraic method.
    '''
    # Make sure the supplied alphabet is kosher. It must contain only single-
    # character strings or `ANYTHING_ELSE`.
    for symbol in f.alphabet:
        if symbol == ANYTHING_ELSE:
            continue
        if isinstance(symbol, str) and len(symbol) == 1:
            continue
        raise Exception(
            f"Symbol {repr(symbol)} cannot be used in a regular expression"
        )

    # We need a new state not already used
    outside = object()

    # The set of strings that would be accepted by this FSM if you started
    # at state i is represented by the regex R_i.
    # If state i has a sole transition "a" to state j, then we know
    # R_i = a R_j.
    # If state i is final, then the empty string is also accepted by this
    # regex.
    # And so on...

    # From this we can build a set of simultaneous equations in len(f.states)
    # variables. This system is easily solved for all variables, but we only
    # need one: R_a, where a is the starting state.

    # The first thing we need to do is organise the states into order of depth,
    # so that when we perform our back-substitutions, we can start with the
    # last (deepest) state and therefore finish with R_a.
    states = [f.initial]
    i = 0
    while i < len(states):
        current = states[i]
        if current in f.map:
            for symbol in sorted(f.map[current], key=alphabet_key):
                next = f.map[current][symbol]
                if next not in states:
                    states.append(next)
        i += 1

    # Our system of equations is represented like so:
    brz = {}
    for a in f.states:
        brz[a] = {}
        for b in f.states:
            brz[a][b] = NULLPATTERN

        if a in f.finals:
            brz[a][outside] = Pattern(EMPTYSTRING)
        else:
            brz[a][outside] = NULLPATTERN

    # Populate it with some initial data.
    for a in f.map:
        for symbol in f.map[a]:
            b = f.map[a][symbol]
            if symbol == ANYTHING_ELSE:
                charclass = ~Charclass(f.alphabet - {ANYTHING_ELSE})
            else:
                charclass = Charclass({symbol})

            brz[a][b] = Pattern(
                *brz[a][b].concs,
                Conc(Mult(charclass, ONE))
            ).reduce()

    # Now perform our back-substitution
    for i in reversed(range(len(states))):
        a = states[i]

        # Before the equation for R_a can be substituted into the other
        # equations, we need to resolve the self-transition (if any).
        # e.g.    R_a = 0 R_a |   1 R_b |   2 R_c
        # becomes R_a =         0*1 R_b | 0*2 R_c
        loop = Mult(brz[a][a], STAR)  # i.e. "0*"
        del brz[a][a]

        for right in brz[a]:
            brz[a][right] = Pattern(
                Conc(
                    loop,
                    Mult(brz[a][right], ONE)
                )
            ).reduce()

        # Note: even if we're down to our final equation, the above step still
        # needs to be performed before anything is returned.

        # Now we can substitute this equation into all of the previous ones.
        for j in range(i):
            b = states[j]

            # e.g. substituting R_a =  0*1 R_b |      0*2 R_c
            # into              R_b =    3 R_a |        4 R_c | 5 R_d
            # yields            R_b = 30*1 R_b | (30*2|4) R_c | 5 R_d
            univ = brz[b][a]  # i.e. "3"
            del brz[b][a]

            for right in brz[a]:
                brz[b][right] = Pattern(
                    *brz[b][right].concs,
                    Conc(
                        Mult(univ, ONE),
                        Mult(brz[a][right], ONE)
                    )
                ).reduce()

    return brz[f.initial][outside].reduce()


def call_fsm(method):
    '''
        Take a method which acts on 0 or more regular expression objects...
        return a new method which simply converts them all to FSMs, calls the
        FSM method on them instead, then converts the result back to a regular
        expression. We do this for several of the more annoying operations.
    '''
    fsm_method = getattr(Fsm, method.__name__)

    def new_method(*elems):
        alphabet = set().union(*[elem.alphabet() for elem in elems])
        return from_fsm(fsm_method(*[elem.to_fsm(alphabet) for elem in elems]))
    return new_method


@dataclass(frozen=True)
class Pattern:
    '''
        A `Pattern` (also known as an "alt", short for "alternation") is a
        set of `Conc`s. A `Pattern` expresses multiple alternate possibilities.
        When written out as a regex, these would separated by pipes. A
        `Pattern` containing no possibilities is possible and represents a
        regular expression matching no strings whatsoever (there is no
        conventional string form for this).

        e.g. "abc|def(ghi|jkl)" is an alt containing two `Conc`s: "abc" and
        "def(ghi|jkl)". The latter is a `Conc` containing four `Mult`s: "d",
        "e", "f" and "(ghi|jkl)". The latter in turn is a `Mult` consisting of
        an upper bound 1, a lower bound 1, and a multiplicand which is a new
        subpattern, "ghi|jkl". This new subpattern again consists of two
        `Conc`s: "ghi" and "jkl".
    '''
    def __init__(self, *concs):
        object.__setattr__(self, "concs", frozenset(concs))

    def __eq__(self, other):
        return isinstance(other, Pattern) and self.concs == other.concs

    def __hash__(self):
        return hash(self.concs)

    def __repr__(self):
        args = ", ".join(repr(conc) for conc in self.concs)
        return f"Pattern({args})"

    def alphabet(self):
        return {ANYTHING_ELSE}.union(*[c.alphabet() for c in self.concs])

    def empty(self):
        return all(conc.empty() for conc in self.concs)

    def intersection(self, other):
        # A deceptively simple method for an astoundingly difficult operation
        alphabet = self.alphabet() | other.alphabet()

        # Which means that we can build finite state machines sharing that
        # alphabet
        combined = self.to_fsm(alphabet) & other.to_fsm(alphabet)
        return from_fsm(combined)

    def __and__(self, other):
        return self.intersection(other)

    @call_fsm
    def difference(*elems):
        '''
            Return a regular expression which matches any string which `self`
            matches but none of the strings which `other` matches.
        '''
        pass

    def __sub__(self, other):
        return self.difference(other)

    def union(self, other):
        return Pattern(*(self.concs | other.concs))

    def __or__(self, other):
        return self.union(other)

    def __str__(self):
        if len(self.concs) == 0:
            raise Exception(f"Can't serialise {repr(self)}")
        return "|".join(sorted(str(conc) for conc in self.concs))

    def reduce(self):
        if self == NULLPATTERN:
            return self

        if self.empty():
            return NULLPATTERN

        # Try recursively reducing our internal `Conc`s.
        reduced = frozenset(c.reduce() for c in self.concs)
        if reduced != self.concs:
            return Pattern(*reduced).reduce()

        # If one of our internal concs is empty, remove it
        for conc in self.concs:
            if conc.empty():
                new = self.concs - {conc}
                return Pattern(*new).reduce()

        # If we have just one `Conc` with just one `Mult` with a multiplier of
        # 1, and the multiplicand is a `Pattern`, pull that up
        if len(self.concs) == 1:
            (conc,) = self.concs
            if len(conc.mults) == 1 \
               and conc.mults[0].multiplier == ONE \
               and isinstance(conc.mults[0].multiplicand, Pattern):
                return conc.mults[0].multiplicand.reduce()

        # If this `Pattern` contains several `Conc`s each containing just 1
        # `Mult` and their multiplicands agree, we may be able to merge the
        # multipliers.
        # e.g. "a{1,2}|a{3,4}|bc" -> "a{1,4}|bc"
        oldconcs = list(self.concs)  # so we can index the things
        for i in range(len(oldconcs)):
            conc1 = oldconcs[i]
            if len(conc1.mults) != 1:
                continue
            multiplicand1 = conc1.mults[0].multiplicand
            for j in range(i + 1, len(oldconcs)):
                conc2 = oldconcs[j]
                if len(conc2.mults) != 1:
                    continue
                multiplicand2 = conc2.mults[0].multiplicand
                if multiplicand2 != multiplicand1:
                    continue
                multiplicand = multiplicand1
                multiplier1 = conc1.mults[0].multiplier
                multiplier2 = conc2.mults[0].multiplier
                if not multiplier1.canunion(multiplier2):
                    continue
                multiplier = multiplier1 | multiplier2
                newconcs = \
                    oldconcs[:i] + \
                    oldconcs[i + 1:j] + \
                    oldconcs[j + 1:] + \
                    [Conc(Mult(multiplicand, multiplier))]
                return Pattern(*newconcs).reduce()

        # If this `Pattern` contains several `Conc`s each containing just 1
        # `Mult` each containing just a `Charclass`, with a multiplier of 1,
        # then we can merge those `Charclass`es together.
        # e.g. "0|[1-9]|ab" -> "[0-9]|ab"
        merged_charclass = NULLCHARCLASS
        num_merged = 0
        rest = []
        for conc in self.concs:
            if len(conc.mults) == 1 \
               and conc.mults[0].multiplier == ONE \
               and isinstance(conc.mults[0].multiplicand, Charclass):
                merged_charclass |= conc.mults[0].multiplicand
                num_merged += 1
            else:
                rest.append(conc)
        if num_merged >= 2:
            rest.append(Conc(Mult(merged_charclass, ONE)))
            return Pattern(*rest).reduce()

        # If one of the present `Pattern`'s `Conc`s is the empty string...
        if EMPTYSTRING in self.concs:
            for conc in self.concs:
                # ...and there is another `Conc`
                # with a single `Mult` whose lower bound is 0...
                if len(conc.mults) == 1 \
                   and conc.mults[0].multiplier.min == Bound(0):
                    # Then we can omit the empty string.
                    # E.g. "|(ab)*|def" => "(ab)*|def".
                    rest = self.concs - {EMPTYSTRING}
                    return Pattern(*rest).reduce()

            for conc in self.concs:
                # ...and there is another `Conc`
                # with a single `Mult` whose lower bound is 1...
                if len(conc.mults) == 1 \
                   and conc.mults[0].multiplier.min == Bound(1):
                    # Then we can merge the empty string into that.
                    # E.g. "|(ab)+|def" => "(ab)*|def".
                    merged_conc = Conc(
                        Mult(
                            conc.mults[0].multiplicand,
                            conc.mults[0].multiplier * QM
                        )
                    )
                    rest = self.concs - {EMPTYSTRING, conc} | {merged_conc}
                    return Pattern(*rest).reduce()

        # If the present `Pattern`'s `Conc`s all have a common prefix, split
        # that out. This increases the depth of the object
        # but it is still arguably simpler/ripe for further reduction
        # e.g. "abc|ade" -> a(bc|de)"
        if len(self.concs) > 1:
            prefix = self._commonconc()
            if prefix != EMPTYSTRING:
                leftovers = self.behead(prefix)
                mults = prefix.mults + (Mult(leftovers, ONE),)
                return Pattern(Conc(*mults)).reduce()

            # Same but for suffixes.
            # e.g. "xyz|stz -> (xy|st)z"
            suffix = self._commonconc(suffix=True)
            if suffix != EMPTYSTRING:
                leftovers = self.dock(suffix)
                mults = (Mult(leftovers, ONE),) + suffix.mults
                return Pattern(Conc(*mults)).reduce()

        return self

    @call_fsm
    def symmetric_difference(*elems):
        '''
            Return a regular expression matching only the strings recognised by
            `self` or `other` but not both.
        '''
        pass

    def __xor__(self, other):
        return self.symmetric_difference(other)

    def dock(self, other):
        '''
            The opposite of concatenation. Remove a common suffix from the
            present `Pattern`; that is, from each of its constituent concs.

            AYZ|BYZ|CYZ - YZ -> A|B|C.
        '''
        return Pattern(*[conc.dock(other) for conc in self.concs])

    def behead(self, other):
        '''
            Like dock() but the other way around. Remove a common prefix from
            the present `Pattern`; that is, from each of its constituent concs.

            ZA|ZB|ZC.behead(Z) -> A|B|C
        '''
        return Pattern(*[conc.behead(other) for conc in self.concs])

    def _commonconc(self, suffix=False):
        '''
            Find the longest `Conc` which acts as prefix to every `Conc` in
            this `Pattern`. This could be `EMPTYSTRING`. Return the common
            prefix along with all the leftovers after truncating that common
            prefix from each `Conc`.

            "ZA|ZB|ZC" -> "Z", "(A|B|C)"
            "ZA|ZB|ZC|Z" -> "Z", "(A|B|C|)"
            "CZ|CZ" -> "CZ", "()"

            If "suffix" is True, the same result but for suffixes.
        '''
        if len(self.concs) == 0:
            raise Exception(f"Can't call _commonconc on {repr(self)}")

        from functools import reduce
        return reduce(
            lambda x, y: x.common(y, suffix=suffix),
            self.concs
        )

    def to_fsm(self, alphabet=None):
        if alphabet is None:
            alphabet = self.alphabet()

        fsm1 = null(alphabet)
        for conc in self.concs:
            fsm1 |= conc.to_fsm(alphabet)
        return fsm1

    def reversed(self):
        return Pattern(*(c.reversed() for c in self.concs))

    def copy(self):
        '''
            For completeness only, since `set.copy()` also exists. `Pattern`s
            are immutable, so I can see only very odd reasons to need this
        '''
        return Pattern(*self.concs)

    def equivalent(self, other):
        '''
            Two `Pattern`s are equivalent if they recognise the same strings.
            Note that in the general case this is actually quite an intensive
            calculation, but far from unsolvable, as we demonstrate here:
        '''
        return self.to_fsm().equivalent(other.to_fsm())

    def times(self, multiplier):
        '''
            Equivalent to repeated concatenation. Multiplier consists of a
            minimum and a maximum; maximum may be infinite (for Kleene star
            closure). Call using "a = b * qm"
        '''
        return Pattern(Conc(Mult(self, multiplier)))

    def __mul__(self, multiplier):
        return self.times(multiplier)

    @call_fsm
    def everythingbut(self):
        '''
            Return a `Pattern` which will match any string not matched by
            `self`, and which will not match any string matched by `self`.
            Another task which is very difficult in general (and typically
            returns utter garbage when actually printed), but becomes trivial
            to code thanks to FSM routines.
        '''
        pass

    def derive(self, string):
        return from_fsm(self.to_fsm().derive(string))

    @call_fsm
    def isdisjoint(self, other):
        '''
            Treat `self` and `other` as sets of strings and see if they are
            disjoint
        '''
        pass

    def matches(self, string):
        return self.to_fsm().accepts(string)

    def __contains__(self, string):
        '''
            This lets you use the syntax `"a" in pattern` to see whether the
            string "a" is in the set of strings matched by `pattern`.
        '''
        return self.matches(string)

    def __reversed__(self):
        return self.reversed()

    def cardinality(self):
        '''
            Consider the regular expression as a set of strings and return the
            cardinality of that set, or raise an OverflowError if there are
            infinitely many.
        '''
        # There is no way to do this other than converting to an FSM, because
        # the `Pattern` may allow duplicate routes, such as "a|a".
        return self.to_fsm().cardinality()

    def __len__(self):
        return self.cardinality()

    def strings(self, otherchar=None):
        '''
            Each time next() is called on this iterator, a new string is
            returned which this `Pattern` can match. `StopIteration`
            is raised once all such strings have been returned, although a
            regex with a * in may match infinitely many strings.
        '''

        # In the case of a regex like "[^abc]", there are infinitely many
        # (well, a very large finite number of) single characters which will
        # match. It's not productive to iterate over all of these giving every
        # single example. You must supply your own "otherchar" to stand in for
        # all of these possibilities.
        for string in self.to_fsm().strings():

            # Have to represent `ANYTHING_ELSE` somehow.
            if ANYTHING_ELSE in string:
                if otherchar is None:
                    raise Exception("Please choose an `otherchar`")
                string = [
                    otherchar if char == ANYTHING_ELSE else char
                    for char in string
                ]

            yield "".join(string)

    def __iter__(self):
        '''
            This allows you to do `for string in pattern` as a list
            comprehension!
        '''
        return self.strings()


@dataclass(frozen=True)
class Mult:
    '''
        A `Mult` is a combination of a multiplicand with a multiplier (a min
        and a max). The vast majority of characters in regular expressions
        occur without a specific multiplier, which is implicitly equivalent to
        a min of 1 and a max of 1, but many more have explicit multipliers like
        "*" (min = 0, max = INF) and so on.

        e.g. a, b{2}, c?, d*, [efg]{2,5}, f{2,}, (anysubpattern)+, .*, ...
    '''
    multiplicand: Union[Charclass, Pattern]
    multiplier: Multiplier

    def __eq__(self, other):
        return isinstance(other.multiplicand, type(self.multiplicand)) \
               and self.multiplicand == other.multiplicand \
               and self.multiplier == other.multiplier

    def __hash__(self):
        return hash((self.multiplicand, self.multiplier))

    def __repr__(self):
        return f"Mult({repr(self.multiplicand)}, {repr(self.multiplier)})"

    def dock(self, other):
        '''
            "Dock" another `Mult` from this one (i.e. remove part of the tail)
            and return the result. The reverse of concatenation. This is a lot
            trickier.
            e.g. a{4,5} - a{3} = a{1,2}
        '''
        if other.multiplicand != self.multiplicand:
            raise Exception(
                f"Can't subtract {repr(other)} from {repr(self)}"
            )
        return Mult(self.multiplicand, self.multiplier - other.multiplier)

    def common(self, other):
        '''
            Return the common part of these two mults. This is the largest
            `Mult` which can be safely subtracted from both the originals. The
            multiplier on this `Mult` could be `ZERO`: this is the case if, for
            example, the multiplicands disagree.
        '''
        if self.multiplicand == other.multiplicand:
            return Mult(
                self.multiplicand,
                self.multiplier.common(other.multiplier)
            )

        # Multiplicands disagree, no common part at all.
        return Mult(NULLCHARCLASS, ZERO)

    def alphabet(self):
        return {ANYTHING_ELSE} | self.multiplicand.alphabet()

    def empty(self):
        return self.multiplicand.empty() and self.multiplier.min > Bound(0)

    def reduce(self):
        if self == NULLMULT:
            return self

        # Can't match anything: reduce to empty `Mult`
        if self.empty():
            return NULLMULT

        # Try recursively reducing our multiplicand
        reduced = self.multiplicand.reduce()
        if reduced != self.multiplicand:
            return Mult(reduced, self.multiplier).reduce()

        # If our multiplicand is a `Pattern` containing an empty `Conc`
        # we can pull that "optional" bit out into our own multiplier
        # instead.
        # e.g. (A|B|C|) -> (A|B|C)?
        # e.g. (A|B|C|){2} -> (A|B|C){0,2}
        if isinstance(self.multiplicand, Pattern) \
           and EMPTYSTRING in self.multiplicand.concs \
           and self.multiplier.canmultiplyby(QM):
            return Mult(
                Pattern(
                    *filter(
                        lambda conc: len(conc.mults) != 0,
                        self.multiplicand.concs
                    )
                ),
                self.multiplier * QM,
            ).reduce()

        # If our multiplicand is a `Pattern` containing a single `Conc`
        # containing a single `Mult`, we can scrap the `Pattern` in favour of
        # that `Mult`'s multiplicand
        # e.g. ([ab])* -> [ab]*
        # e.g. ((a))* -> (a)* -> a*
        # NOTE: this logic lives here at the `Mult` level, NOT in
        # `Pattern.reduce` because we want to return another `Mult` (same type)
        if isinstance(self.multiplicand, Pattern) \
           and len(self.multiplicand.concs) == 1:
            (conc,) = self.multiplicand.concs
            if len(conc.mults) == 1 \
               and conc.mults[0].multiplier.canmultiplyby(self.multiplier):
                return Mult(
                    conc.mults[0].multiplicand,
                    conc.mults[0].multiplier * self.multiplier
                ).reduce()

        # no reduction possible
        return self

    def __str__(self):
        if isinstance(self.multiplicand, Pattern):
            return f"({str(self.multiplicand)}){str(self.multiplier)}"
        if isinstance(self.multiplicand, Charclass):
            return f"{str(self.multiplicand)}{str(self.multiplier)}"
        raise Exception(f"Unknown type {str(type(self.inner))}")

    def to_fsm(self, alphabet=None):
        if alphabet is None:
            alphabet = self.alphabet()

        # worked example: (min, max) = (5, 7) or (5, INF)
        # (mandatory, optional) = (5, 2) or (5, INF)

        unit = self.multiplicand.to_fsm(alphabet)
        # accepts e.g. "ab"

        # accepts "ababababab"
        mandatory = unit * self.multiplier.mandatory.v

        # unlimited additional copies
        if self.multiplier.optional == INF:
            optional = unit.star()
            # accepts "(ab)*"

        else:
            optional = epsilon(alphabet) | unit
            # accepts "(ab)?"

            optional *= self.multiplier.optional.v
            # accepts "(ab)?(ab)?"

        return mandatory + optional

    def reversed(self):
        return Mult(self.multiplicand.reversed(), self.multiplier)


NULLMULT = Mult(NULLCHARCLASS, ONE)
NULLCONC = Conc(NULLMULT)
EMPTYSTRING = Conc()
NULLPATTERN = Pattern(NULLCONC)



# -*- coding: utf-8 -*-

from .charclass import Charclass, WORDCHAR, DIGIT, SPACECHAR, NONWORDCHAR, \
    NONDIGITCHAR, NONSPACECHAR, DOT, NULLCHARCLASS
from .fsm import ANYTHING_ELSE


def test_charclass_equality():
    assert Charclass("a") == Charclass("a")
    assert ~Charclass("a") == ~Charclass("a")
    assert ~Charclass("a") != Charclass("a")
    assert Charclass("ab") == Charclass("ba")


def test_repr():
    assert repr(~Charclass("a")) == "~Charclass('a')"


def test_charclass_str():
    assert str(WORDCHAR) == "\\w"
    assert str(DIGIT) == "\\d"
    assert str(SPACECHAR) == "\\s"
    assert str(Charclass("a")) == "a"
    assert str(Charclass("{")) == "\\{"
    assert str(Charclass("\t")) == "\\t"
    assert str(Charclass("ab")) == "[ab]"
    assert str(Charclass("a{")) == "[a{]"
    assert str(Charclass("a\t")) == "[\\ta]"
    assert str(Charclass("a-")) == "[\\-a]"
    assert str(Charclass("a[")) == "[\\[a]"
    assert str(Charclass("a]")) == "[\\]a]"
    assert str(Charclass("ab")) == "[ab]"
    assert str(Charclass("abc")) == "[abc]"
    assert str(Charclass("abcd")) == "[a-d]"
    assert str(Charclass("abcdfghi")) == "[a-df-i]"
    assert str(Charclass("^")) == "^"
    assert str(Charclass("\\")) == "\\\\"
    assert str(Charclass("a^")) == "[\\^a]"
    assert str(Charclass("0123456789a")) == "[0-9a]"
    assert str(Charclass("\t\v\r A")) == "[\\t\\v\\r A]"
    assert str(Charclass("\n\f A")) == "[\\n\\f A]"
    assert str(Charclass("\t\n\v\f\r A")) == "[\\t-\\r A]"
    assert str(Charclass(
        "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz|"
    )) == "[0-9A-Z_a-z|]"
    assert str(NONWORDCHAR) == "\\W"
    assert str(NONDIGITCHAR) == "\\D"
    assert str(NONSPACECHAR) == "\\S"
    assert str(DOT) == "."
    assert str(~Charclass("")) == "."
    assert str(~Charclass("a")) == "[^a]"
    assert str(~Charclass("{")) == "[^{]"
    assert str(~Charclass("\t")) == "[^\\t]"
    assert str(~Charclass("^")) == "[^\\^]"


def test_charclass_fsm():
    # "[^a]"
    nota = (~Charclass("a")).to_fsm()
    assert nota.alphabet == {"a", ANYTHING_ELSE}
    assert nota.accepts("b")
    assert nota.accepts(["b"])
    assert nota.accepts([ANYTHING_ELSE])


def test_charclass_negation():
    assert ~~Charclass("a") == Charclass("a")
    assert Charclass("a") == ~~Charclass("a")


def test_empty():
    assert NULLCHARCLASS.empty()
    assert not DOT.empty()



# -*- coding: utf-8 -*-

if __name__ == "__main__":
    raise Exception(
        "Test files can't be run directly. Use `python -m pytest greenery`"
    )

import pickle

from .fsm import Fsm, ANYTHING_ELSE
from .rxelems import from_fsm
from .charclass import DIGIT, WORDCHAR
from .parse import parse


###############################################################################
# Stringification tests

def test_charclass_str():
    # Arbitrary ranges
    assert str(parse("[\\w:;<=>?@\\[\\\\\\]\\^`]")) == "[0-z]"
    # TODO: what if \d is a proper subset of `chars`?

    # escape sequences are not preserved
    assert str(parse("\\x09")) == "\\t"

    # Printing ASCII control characters? You should get hex escapes
    assert str(parse("\\x00")) == "\\x00"


def test_parse_str_round_trip():
    # not "a[ab]b"
    assert str(parse("a.b")) == "a.b"
    assert str(parse("\\d{4}")) == "\\d{4}"
    assert str(parse("a.b()()")) == "a.b()()"


###############################################################################
# Test to_fsm() and alphabet-related functionality

def test_alphabet():
    # `.alphabet()` should include `ANYTHING_ELSE`
    assert parse("").alphabet() == {ANYTHING_ELSE}


def test_pattern_fsm():
    # "a[^a]"
    anota = parse("a[^a]").to_fsm()
    assert len(anota.states) == 3
    assert not anota.accepts("a")
    assert not anota.accepts(["a"])
    assert not anota.accepts("b")
    assert not anota.accepts(["b"])
    assert not anota.accepts("aa")
    assert not anota.accepts(["a", "a"])
    assert anota.accepts("ab")
    assert anota.accepts(["a", "b"])
    assert anota.accepts(["a", ANYTHING_ELSE])
    assert not anota.accepts("ba")
    assert not anota.accepts("bb")

    # "0\\d"
    zeroD = parse("0\\d").to_fsm(DIGIT.chars)
    assert zeroD.accepts("01")
    assert not zeroD.accepts("10")

    # "\\d{2}"
    d2 = parse("\\d{2}").to_fsm(DIGIT.chars)
    assert not d2.accepts("")
    assert not d2.accepts("1")
    assert d2.accepts("11")
    assert not d2.accepts("111")

    # abc|def(ghi|jkl)
    conventional = parse("abc|def(ghi|jkl)").to_fsm(WORDCHAR.chars)
    assert not conventional.accepts("a")
    assert not conventional.accepts("ab")
    assert conventional.accepts("abc")
    assert not conventional.accepts("abcj")
    assert conventional.accepts("defghi")
    assert conventional.accepts("defjkl")


def test_fsm():
    # You should be able to to_fsm() a single regular expression element
    # without supplying a specific alphabet. That should be determinable from
    # context.
    assert parse("a.b").to_fsm().accepts("acb")

    bad = parse("0{2}|1{2}").to_fsm({"0", "1", ANYTHING_ELSE})
    assert bad.accepts("00")
    assert bad.accepts("11")
    assert not bad.accepts("01")

    bad = parse("0{2}|1{2}").to_fsm()
    assert bad.accepts("00")
    assert bad.accepts("11")
    assert not bad.accepts("01")


def test_bug_28():
    # Starification was broken in FSMs
    assert not parse("(ab*)").to_fsm().star().accepts("bb")
    assert not parse("(ab*)*").to_fsm().accepts("bb")


###############################################################################
# Test matches(). Quite sparse at the moment

def test_wildcards_in_charclasses():
    # Allow "\w", "\d" and "\s" in `Charclass`es
    assert parse("[\\w~]*").matches("a0~")
    assert parse("[\\da]*").matches("0129a")
    assert parse("[\\s]+").matches(" \t \t ")


def test_block_comment_regex():
    # I went through several incorrect regexes for C block comments. Here we
    # show why the first few attempts were incorrect
    a = parse("/\\*(([^*]|\\*+[^*/])*)\\*/")
    assert a.matches("/**/")
    assert not a.matches("/***/")
    assert not a.matches("/****/")

    b = parse("/\\*(([^*]|\\*[^/])*)\\*/")
    assert b.matches("/**/")
    assert not b.matches("/***/")
    assert b.matches("/****/")

    c = parse("/\\*(([^*]|\\*+[^*/])*)\\*+/")
    assert c.matches("/**/")
    assert c.matches("/***/")
    assert c.matches("/****/")


def test_named_groups():
    a = parse("(?P<ng1>abc)")
    assert a.matches("abc")


def test_in():
    assert "a" in parse("a")
    assert "abcdsasda" in parse("\\w{4,10}")
    assert "abc" in parse("abc|def(ghi|jkl)")


###############################################################################
# Test string generators

def test_charclass_gen():
    gen = parse("[xyz]").strings()
    assert next(gen) == "x"
    assert next(gen) == "y"
    assert next(gen) == "z"
    try:
        next(gen)
        assert False
    except StopIteration:
        assert True


def test_mult_gen():
    # One term
    gen = parse("[ab]").strings()
    assert next(gen) == "a"
    assert next(gen) == "b"
    try:
        next(gen)
        assert False
    except StopIteration:
        assert True

    # No terms
    gen = parse("[ab]{0}").strings()
    assert next(gen) == ""
    try:
        next(gen)
        assert False
    except StopIteration:
        assert True

    # Many terms
    gen = parse("[ab]*").strings()
    assert next(gen) == ""
    assert next(gen) == "a"
    assert next(gen) == "b"
    assert next(gen) == "aa"
    assert next(gen) == "ab"
    assert next(gen) == "ba"
    assert next(gen) == "bb"
    assert next(gen) == "aaa"


def test_conc_generator():
    gen = parse("[ab][cd]").strings()
    assert next(gen) == "ac"
    assert next(gen) == "ad"
    assert next(gen) == "bc"
    assert next(gen) == "bd"
    try:
        next(gen)
        assert False
    except StopIteration:
        assert True


def test_pattern_generator():
    gen = parse("[ab]|[cde]").strings()
    assert next(gen) == "a"
    assert next(gen) == "b"
    assert next(gen) == "c"
    assert next(gen) == "d"
    assert next(gen) == "e"
    try:
        next(gen)
        assert False
    except StopIteration:
        assert True

    # more complex
    gen = parse("abc|def(ghi|jkl)").strings()
    assert next(gen) == "abc"
    assert next(gen) == "defghi"
    assert next(gen) == "defjkl"

    gen = parse("[0-9a-fA-F]{3,10}").strings()
    assert next(gen) == "000"
    assert next(gen) == "001"
    assert next(gen) == "002"


def test_infinite_generation():
    # Infinite generator, flummoxes both depth-first and breadth-first searches
    gen = parse("a*b*").strings()
    assert next(gen) == ""
    assert next(gen) == "a"
    assert next(gen) == "b"
    assert next(gen) == "aa"
    assert next(gen) == "ab"
    assert next(gen) == "bb"
    assert next(gen) == "aaa"
    assert next(gen) == "aab"
    assert next(gen) == "abb"
    assert next(gen) == "bbb"
    assert next(gen) == "aaaa"


def test_wildcard_generator():
    # Generator needs to handle wildcards as well. Wildcards come last.
    gen = parse("a.b").strings(otherchar="*")
    assert next(gen) == "aab"
    assert next(gen) == "abb"
    assert next(gen) == "a*b"
    try:
        next(gen)
        assert False
    except StopIteration:
        assert True


def test_forin():
    assert [s for s in parse("abc|def(ghi|jkl)")] \
        == ["abc", "defghi", "defjkl"]


###############################################################################
# Test cardinality() and len()

def test_cardinality():
    assert parse("[]").cardinality() == 0
    assert parse("[]?").cardinality() == 1
    assert parse("[]{0,6}").cardinality() == 1
    assert parse("[ab]{3}").cardinality() == 8
    assert parse("[ab]{2,3}").cardinality() == 12
    assert len(parse("abc|def(ghi|jkl)")) == 3
    try:
        len(parse(".*"))
        assert False
    except OverflowError:
        assert True


###############################################################################

def test_copy():
    x = parse("abc|def(ghi|jkl)")
    assert x.copy() == x


###############################################################################
# Test from_fsm()

def test_dot():
    # not "a[ab]b"
    assert str(from_fsm(parse("a.b").to_fsm())) == "a.b"


def test_abstar():
    # Buggggs.
    abstar = Fsm(
        alphabet={'a', ANYTHING_ELSE, 'b'},
        states={0, 1},
        initial=0,
        finals={0},
        map={
            0: {'a': 0, ANYTHING_ELSE: 1, 'b': 0},
            1: {'a': 1, ANYTHING_ELSE: 1, 'b': 1}
        }
    )
    assert str(from_fsm(abstar)) == "[ab]*"


def test_adotb():
    adotb = Fsm(
        alphabet={'a', ANYTHING_ELSE, 'b'},
        states={0, 1, 2, 3, 4},
        initial=0,
        finals={4},
        map={
            0: {'a': 2, ANYTHING_ELSE: 1, 'b': 1},
            1: {'a': 1, ANYTHING_ELSE: 1, 'b': 1},
            2: {'a': 3, ANYTHING_ELSE: 3, 'b': 3},
            3: {'a': 1, ANYTHING_ELSE: 1, 'b': 4},
            4: {'a': 1, ANYTHING_ELSE: 1, 'b': 1}
        }
    )
    assert str(from_fsm(adotb)) == "a.b"


def test_rxelems_recursion_error():
    # Catch a recursion error
    assert str(from_fsm(Fsm(
        alphabet={"0", "1"},
        states={0, 1, 2, 3},
        initial=3,
        finals={1},
        map={
            0: {"0": 1, "1": 1},
            1: {"0": 2, "1": 2},
            2: {"0": 2, "1": 2},
            3: {"0": 0, "1": 2},
        }
    ))) == "0[01]"


def test_even_star_bug1():
    # Bug fix. This is a(a{2})* (i.e. accepts an odd number of "a" chars in a
    # row), but when from_fsm() is called, the result is "a+". Turned out to be
    # a fault in the rxelems.multiplier.__mul__() routine
    elesscomplex = Fsm(
        alphabet={"a"},
        states={0, 1},
        initial=0,
        finals={1},
        map={
            0: {"a": 1},
            1: {"a": 0},
        },
    )
    assert not elesscomplex.accepts("")
    assert elesscomplex.accepts("a")
    assert not elesscomplex.accepts("aa")
    assert elesscomplex.accepts("aaa")
    elesscomplex = from_fsm(elesscomplex)
    assert str(elesscomplex) in {"a(a{2})*", "(a{2})*a"}
    elesscomplex = elesscomplex.to_fsm()
    assert not elesscomplex.accepts("")
    assert elesscomplex.accepts("a")
    assert not elesscomplex.accepts("aa")
    assert elesscomplex.accepts("aaa")
    gen = elesscomplex.strings()
    assert next(gen) == ["a"]
    assert next(gen) == ["a", "a", "a"]
    assert next(gen) == ["a", "a", "a", "a", "a"]
    assert next(gen) == ["a", "a", "a", "a", "a", "a", "a"]


def test_binary_3():
    # Binary numbers divisible by 3.
    # Disallows the empty string
    # Allows "0" on its own, but not leading zeroes.
    div3 = from_fsm(Fsm(
        alphabet={"0", "1"},
        states={"initial", "zero", 0, 1, 2, None},
        initial="initial",
        finals={"zero", 0},
        map={
            "initial": {"0": "zero", "1": 1},
            "zero": {"0": None, "1": None},
            0: {"0": 0, "1": 1},
            1: {"0": 2, "1": 0},
            2: {"0": 1, "1": 2},
            None: {"0": None, "1": None},
        },
    ))
    assert str(parse("(0|1)").reduce()) == "[01]"
    assert str(parse("(0|12)").reduce()) == "0|12"
    assert str(parse("(0|1(01*0|10*1)*10*)").reduce()) == "0|1(01*0|10*1)*10*"
    assert str(div3) == "0|1(01*0|10*1)*10*"
    gen = div3.strings()
    assert next(gen) == "0"
    assert next(gen) == "11"
    assert next(gen) == "110"
    assert next(gen) == "1001"
    assert next(gen) == "1100"


def test_base_N():
    # Machine accepts only numbers in selected base (e.g. 2, 10) that are
    # divisible by N (e.g. 3, 7).
    # "0" alone is acceptable, but leading zeroes (e.g. "00", "07") are not
    base = 2
    N = 3
    assert base <= 10
    divN = from_fsm(Fsm(
        alphabet=set(str(i) for i in range(base)),
        states=set(range(N)) | {"initial", "zero", None},
        initial="initial",
        finals={"zero", 0},
        map=dict(
            [
                (
                    "initial",
                    dict(
                        [(str(j), j % N) for j in range(1, base)] +
                        [("0", "zero")]
                    )
                ),
                (
                    "zero",
                    dict(
                        [(str(j), None) for j in range(base)]
                    )
                ),
                (
                    None,
                    dict(
                        [(str(j), None) for j in range(base)]
                    )
                ),
            ] + [
                (
                    i,
                    dict(
                        [(str(j), (i * base + j) % N) for j in range(base)]
                    )
                )
                for i in range(N)
            ]
        ),
    ))
    gen = divN.strings()
    a = next(gen)
    assert a == "0"
    for i in range(7):
        b = next(gen)
        assert int(a, base) + N == int(b, base)
        a = b


def test_bad_alphabet():
    # You can use anything you like in your FSM alphabet, but if you try to
    # convert it to an `rxelems` object then the only acceptable symbols are
    # single characters or `ANYTHING_ELSE`.
    for bad_symbol in [None, (), 0, ("a",), "", "aa", "ab", True]:
        f = Fsm(
            alphabet={bad_symbol},
            states={0},
            initial=0,
            finals=set(),
            map={
                0: {bad_symbol: 0}
            },
        )
        try:
            from_fsm(f)
            assert False
        except AssertionError:
            raise Exception(f"Accepted bad symbol: {repr(bad_symbol)}")
        except Exception:
            pass


def test_dead_default():
    blockquote = from_fsm(Fsm(
        alphabet={"/", "*", ANYTHING_ELSE},
        states={0, 1, 2, 3, 4},
        initial=0,
        finals={4},
        map={
                0: {"/": 1},
                1: {"*": 2},
                2: {"/": 2, ANYTHING_ELSE: 2, "*": 3},
                3: {"/": 4, ANYTHING_ELSE: 2, "*": 3},
        }
    ))
    assert str(blockquote) == "/\\*([^*]|\\*+[^*/])*\\*+/"


###############################################################################
# charclass set operations

def test_charclass_union():
    assert (parse("[ab]") | parse("[bc]")).reduce() == parse("[abc]")
    assert (parse("[ab]") | parse("[^bc]")).reduce() == parse("[^c]")
    assert (parse("[^ab]") | parse("[bc]")).reduce() == parse("[^a]")
    assert (parse("[^ab]") | parse("[^bc]")).reduce() == parse("[^b]")


def test_charclass_intersection():
    assert parse("[ab]") & parse("[bc]") == parse("b")
    assert parse("[ab]") & parse("[^bc]") == parse("a")
    assert parse("[^ab]") & parse("[bc]") == parse("c")
    assert parse("[^ab]") & parse("[^bc]") == parse("[^abc]")


###############################################################################
# Emptiness detection

def test_empty():
    assert not parse("a{0}").empty()
    assert parse("[]").empty()
    assert not parse("[]?").empty()
    assert parse("a[]").empty()
    assert not parse("a[]?").empty()
    assert not parse("a{0}").empty()
    assert not parse("[]?").empty()


###############################################################################
# Test everythingbut()

def test_everythingbut():
    # Regexes are usually gibberish but we make a few claims
    a = parse("a")
    notA = a.everythingbut().to_fsm()
    assert notA.accepts("")
    assert not notA.accepts("a")
    assert notA.accepts("aa")

    # everythingbut(), called twice, should take us back to where we started.
    beer = parse("beer")
    notBeer = beer.everythingbut()
    beer2 = notBeer.everythingbut()
    assert str(beer2) == "be{2}r"

    # ".*" becomes "[]" and vice versa under this call.
    assert str(parse(".*").everythingbut()) == "[]"
    assert str(parse("[]").everythingbut()) == ".*"


def test_isinstance_bug():
    # Problem relating to isinstance(). The class `Mult` was occurring as both
    # rxelems.Mult and as __main__.Mult and apparently these count as different
    # classes for some reason, so isinstance(m, Mult) was returning false.
    var = str(parse("").everythingbut()) \
        + "aa" \
        + str(parse("").everythingbut())
    assert var == ".+aa.+"

    starfree = parse(var).everythingbut()
    assert str(starfree) == "(.(a?[^a])*a{0,2})?"


###############################################################################

def test_equivalence():
    assert parse("aa*").equivalent(parse("a*a"))
    assert parse("([ab]*a|[bc]*c)?b*").equivalent(parse("b*(a[ab]*|c[bc]*)?"))


###############################################################################
# Test reversed()

def test_regex_reversal():
    assert reversed(parse("b")) == parse("b")
    assert reversed(parse("e*")) == parse("e*")
    assert reversed(parse("bear")) == parse("raeb")
    assert reversed(parse("beer")) == parse("reeb")
    assert reversed(parse("abc|def|ghi")) == parse("cba|fed|ihg")
    assert reversed(parse("(abc)*d")) == parse("d(cba)*")


###############################################################################
# Tests for some more set operations

def test_set_ops():
    assert parse("[abcd]") - parse("a") == parse("[bcd]")
    assert parse("[abcd]") ^ parse("[cdef]") == parse("[abef]")


###############################################################################
# Test methods for finding common parts of regular expressions.

def test_pattern_commonconc():
    assert str(parse("aa|aa")._commonconc()) == "aa"
    assert str(parse("abc|aa")._commonconc()) == "a"
    assert str(parse("a|bc")._commonconc()) == ""
    assert str(parse("cf{1,2}|cf")._commonconc()) == "cf"
    assert str(parse("ZA|ZB|ZC")._commonconc()) == "Z"
    assert str(parse("Z+A|ZB|ZZC")._commonconc()) == "Z"
    assert str(parse("a{2}b|a+c")._commonconc()) == "a"


def test_pattern_commonconc_suffix():
    assert str(parse("a|bc")._commonconc(suffix=True)) == ""
    assert str(parse("aa|bca")._commonconc(suffix=True)) == "a"
    assert str(parse("xyza|abca|a")._commonconc(suffix=True)) == "a"
    assert str(parse("f{2,3}c|fc")._commonconc(suffix=True)) == "fc"
    assert str(parse("aa")._commonconc(suffix=True)) == "aa"


###############################################################################
# Basic concatenation reduction tests

def test_reduce_concatenations():
    assert str(parse("aa").reduce()) == "a{2}"
    assert str(parse("bb").reduce()) == "b{2}"
    assert str(parse("b*b").reduce()) == "b+"
    assert str(parse("aa{2,}").reduce()) == "a{3,}"
    assert str(parse("a*a{2}").reduce()) == "a{2,}"
    assert str(parse("aa{0,8}").reduce()) == "a{1,9}"
    assert str(parse("b{0,8}b").reduce()) == "b{1,9}"
    assert str(parse("aab").reduce()) == "a{2}b"
    assert str(parse("abb").reduce()) == "ab{2}"
    assert str(parse("abb*").reduce()) == "ab+"
    assert str(parse("abbc").reduce()) == "ab{2}c"
    assert str(parse("a?ab").reduce()) == "a{1,2}b"
    assert str(parse("(ac{2}|bc+)c").reduce()) == "(ac|bc*)c{2}"
    assert str(parse("a(a{2}b|a+c)").reduce()) == "a{2}(a*c|ab)"
    assert str(parse("a{2,3}(a{2}b|a+c)").reduce()) == "a{3,4}(a*c|ab)"
    assert str(parse("(ba{2}|ca+)a{2,3}").reduce()) == "(ba|ca*)a{3,4}"
    assert str(parse("za{2,3}(a{2}b|a+c)").reduce()) == "za{3,4}(a*c|ab)"
    assert str(parse("(ba{2}|ca+)a{2,3}z").reduce()) == "(ba|ca*)a{3,4}z"
    assert str(parse("(a|bc)(a|bc)").reduce()) == "(a|bc){2}"
    assert str(parse("a+[ab]+").reduce()) == "a[ab]+"
    assert str(parse("a{3,8}[ab]+").reduce()) == "a{3}[ab]+"
    assert str(parse("[ab]+b+").reduce()) == "[ab]+b"
    assert str(parse("[ab]+a{3,8}").reduce()) == "[ab]+a{3}"
    assert str(parse("\\d+\\w+").reduce()) == "\\d\\w+"
    assert str(parse("[ab]+a?").reduce()) == "[ab]+"


###############################################################################
# Multiplication tests

def test_mult_multiplication():
    assert parse("(a{2,3}){1,1}").reduce() == \
        parse("a{2,3}").reduce()
    assert parse("(a{2,3}){1}").reduce() == \
        parse("a{2,3}").reduce()
    assert parse("(a{2,3})").reduce() == \
        parse("a{2,3}").reduce()
    assert parse("(a{2,3}){4,5}").reduce() == \
        parse("a{8,15}").reduce()
    assert parse("(a{2,}){2,}").reduce() == \
        parse("a{4,}").reduce()


def test_even_star_bug2():
    # Defect: (a{2})* should NOT reduce to a*
    assert parse("(a{2})*").reduce() != parse("a*").reduce()


def test_two_or_more_qm_bug():
    assert str(parse("(a{2,})?").reduce()) == "(a{2,})?"


def test_two_two_bug():
    assert str(parse("(a{2}){2}").reduce()) == "a{4}"


###############################################################################
# Test intersection (&)

def test_mult_intersection():
    assert str(parse("a") & parse("a")) == "a"
    assert str(parse("a*") & parse("a")) == "a"
    assert str(parse("a") & parse("a?")) == "a"
    assert str(parse("a{2}") & parse("a{2,}")) == "a{2}"
    assert str(parse("a*") & parse("a+")) == "a+"
    assert str(parse("a{2}") & parse("a{4}")) == "[]"
    assert str(parse("a{3,}") & parse("a{3,}")) == "a{3,}"


def test_parse_regex_intersection():
    assert str(parse("a*") & parse("b*")) == ""
    assert str(parse("a") & parse("b")) == "[]"
    assert str(parse("\\d") & parse(".")) == "\\d"
    assert str(parse("\\d{2}") & parse("0.")) == "0\\d"
    assert str(parse("\\d{2}") & parse("19.*")) == "19"
    assert str(parse("\\d{3}") & parse("19.*")) == "19\\d"
    assert str(parse("abc...") & parse("...def")) == "abcdef"
    assert str(parse("[bc]*[ab]*") & parse("[ab]*[bc]*")) in {
        "([ab]*a|[bc]*c)?b*",
        "b*(a[ab]*|c[bc]*)?"
    }
    assert str(parse("\\W*")) == "\\W*"
    assert str(parse("[a-g0-8$%\\^]+")) == "[$%0-8\\^a-g]+"
    assert str(parse("[^d]{2,8}")) == "[^d]{2,8}"
    assert str(parse("\\W*") & parse("[a-g0-8$%\\^]+")) == "[$%\\^]+"
    assert str(parse("[ab]{1,2}") & parse("[^a]{1,2}")) == "b{1,2}"
    assert str(parse("[ab]?") & parse("[^a]?")) == "b?"
    assert parse("a{0,2}").matches("")
    assert parse("[ab]{0,2}").matches("")
    assert parse("[^a]{0,2}").matches("")
    assert parse("b{0,2}").matches("")
    assert str(parse("[ab]{0,2}") & parse("[^a]{0,2}")) == \
        "b{0,2}"
    assert str(parse("[ab]{0,4}") & parse("[^a]{0,4}")) == \
        "b{0,4}"
    assert str(parse("[abc]{0,8}") & parse("[^a]{0,8}")) == \
        "[bc]{0,8}"
    assert str(parse("[a-g0-8$%\\^]{0,8}") & parse("[^d]{0,8}")) == \
        "[$%0-8\\^abcefg]{0,8}"
    assert str(parse("[a-g0-8$%\\^]+") & parse("[^d]{0,8}")) == \
        "[$%0-8\\^abcefg]{1,8}"
    assert str(parse("[a-g0-8$%\\^]+") & parse("[^d]{2,8}")) == \
        "[$%0-8\\^abcefg]{2,8}"
    assert str(
        parse("\\W*") &
        parse("[a-g0-8$%\\^]+") &
        parse("[^d]{2,8}")
    ) == \
        "[$%\\^]{2,8}"
    assert str(parse("\\d{4}-\\d{2}-\\d{2}") & parse("19.*")) == \
        "19\\d{2}-\\d{2}-\\d{2}"


def test_complexify():
    # Complexify!
    gen = (parse("[bc]*[ab]*") & parse("[ab]*[bc]*")).strings()
    assert next(gen) == ""
    assert next(gen) == "a"
    assert next(gen) == "b"
    assert next(gen) == "c"
    assert next(gen) == "aa"
    assert next(gen) == "ab"
    # no "ac"
    assert next(gen) == "ba"
    assert next(gen) == "bb"
    assert next(gen) == "bc"
    # no "ca"
    assert next(gen) == "cb"
    assert next(gen) == "cc"
    assert next(gen) == "aaa"


def test_silly_reduction():
    # This one is horrendous and we have to jump through some hoops to get to
    # a sensible result. Probably not a good unit test actually.
    long = \
        "(aa|bb*aa)a*|((ab|bb*ab)|(aa|bb*aa)a*b)" + \
        "((ab|bb*ab)|(aa|bb*aa)a*b)*" + \
        "(aa|bb*aa)a*|((ab|bb*ab)|(aa|bb*aa)a*b)" + \
        "((ab|bb*ab)|(aa|bb*aa)a*b)*"
    long = parse(long)
    long = reversed(long.to_fsm())
    long = reversed(from_fsm(long))
    assert str(long) == "[ab]*a[ab]"
    short = "[ab]*a?b*|[ab]*b?a*"
    assert str(parse(".*") & parse(short)) == "[ab]*"


###############################################################################
# reduce() tests

def test_mult_reduction_easy():
    assert str(parse("a").reduce()) == "a"
    assert str(parse("a").reduce()) == "a"
    assert str(parse("a?").reduce()) == "a?"
    assert str(parse("a{0}").reduce()) == ""
    assert str(parse("[]").reduce()) == "[]"
    assert str(parse("[]?").reduce()) == ""
    assert str(parse("[]{0}").reduce()) == ""
    assert str(parse("[]{0,5}").reduce()) == ""


def test_conc_reduction_basic():
    assert str(parse("a").reduce()) == "a"
    assert str(parse("a{3,4}").reduce()) == "a{3,4}"
    assert str(parse("ab").reduce()) == "ab"
    assert str(parse("a[]b").reduce()) == "[]"


def test_pattern_reduce_basic():
    assert str(parse("ab|cd").reduce()) == "ab|cd"
    assert str(parse("a{2}b{2}").reduce()) == "a{2}b{2}"
    assert str(parse("a{2}").reduce()) == "a{2}"
    assert str(parse("a").reduce()) == "a"
    assert str(parse("(((a)))").reduce()) == "a"


def test_empty_conc_suppression():
    assert str(parse("[]0\\d").reduce()) == "[]"


def test_nested_pattern_reduction():
    # a(d(ab|a*c)) -> ad(ab|a*c)
    assert str(parse("a(d(ab|a*c))").reduce()) == "ad(a*c|ab)"


def test_mult_factor_out_qm():
    # `Mult` contains a `Pattern` containing an empty `Conc`? Pull the empty
    # part out where it's external
    assert str(parse("a|b*|").reduce()) == "a|b*"
    assert str(parse("(a|b*|)").reduce()) == "a|b*"
    assert str(parse("(a|b*|)c").reduce()) == "(a|b*)c"
    # This happens even if `EMPTYSTRING` is the only thing left inside the
    # `Mult`
    assert str(parse("()").reduce()) == ""
    assert str(parse("([$%\\^]|){1}").reduce()) == "[$%\\^]?"


def test_remove_unnecessary_parens():
    # `Mult` contains a `Pattern` containing a single `Conc` containing a
    # single `Mult`? That can be reduced greatly
    assert str(parse("(a){2}b").reduce()) == "a{2}b"
    assert str(parse("(a?)+b").reduce()) == "a*b"
    assert str(parse("([ab])*").reduce()) == "[ab]*"
    assert str(parse("(c{1,2}){3,4}").reduce()) == "c{3,8}"


def test_obvious_reduction():
    assert str(parse("(a|b)*").reduce()) == "[ab]*"


def test_mult_squoosh():
    # sequence squooshing of mults within a `Conc`
    assert str(parse("[0-9a-fA-F][0-9a-fA-F][0-9a-fA-F]").reduce()) == \
        "[0-9A-Fa-f]{3}"
    assert str(parse("[$%\\^]?[$%\\^]").reduce()) == \
        "[$%\\^]{1,2}"
    assert str(parse(
        "(|(|(|(|(|[$%\\^])[$%\\^])[$%\\^])[$%\\^])[$%\\^])[$%\\^][$%\\^]"
    ).reduce()) == \
        "[$%\\^]{2,7}"


def test_bad_reduction_bug():
    # DEFECT: "0{2}|1{2}" was erroneously reduced() to "[01]{2}"
    assert parse("0{2}|1{2}").reduce() != \
        parse("[01]{2}")
    assert parse("0|[1-9]|ab").reduce() == \
        parse("\\d|ab")
    assert parse("0|[1-9]|a{5,7}").reduce() == \
        parse("\\d|a{5,7}")
    assert parse("0|(0|[1-9]|a{5,7})").reduce() == \
        parse("0|(\\d|a{5,7})")
    # TODO: should do better than this! Merge that 0


def test_common_prefix_pattern_reduction():
    assert str(parse("a{2}b|a+c").reduce()) == "a(a*c|ab)"


def test_epsilon_reduction():
    assert str(parse("|(ab)*|def").reduce()) == "(ab)*|def"
    assert str(parse("|(ab)+|def").reduce()) == "(ab)*|def"
    assert str(parse("|.+").reduce()) == ".*"
    assert str(parse("|a+|b+").reduce()) in {
        "a+|b*",
        "a*|b+"
    }


def test_charclass_intersection_2():
    assert (parse("[A-z]") & parse("[^g]")).reduce() == \
        parse("[A-fh-z]").reduce()


def test_reduce_boom():
    # make sure recursion problem in reduce() has gone away
    assert str(parse("([1-9]|0)").reduce()) == "\\d"


def test_new_reduce():
    # The @reduce_after decorator has been removed from many methods since it
    # takes unnecessary time which the user may not wish to spend.
    # This alters the behaviour of several methods and also exposes a new
    # opportunity for `Conc.reduce()`
    assert str(parse("a()").reduce()) == "a"
    assert str(parse("a()()").reduce()) == "a"
    assert str(parse("a.b()()").reduce()) == "a.b"


def test_main_bug():
    assert str(parse("a*").reduce()) == \
        "a*"
    assert str(parse("a|a*").reduce()) == \
        "a*"
    assert str(parse("a{1,2}|a{3,4}|bc").reduce()) == \
        "a{1,4}|bc"
    assert str(parse("a{1,2}|bc|a{3,4}").reduce()) == \
        "a{1,4}|bc"
    assert str(parse("a{1,2}|a{3,4}|a{5,6}|bc").reduce()) == \
        "a{1,6}|bc"
    assert str(parse("a{1,2}|a{3}|a{5,6}").reduce()) == \
        "a{1,2}(a?|a{4})"
    assert str(parse("a{1,2}|a{3}|a{5,6}|bc").reduce()) == \
        "a{1,3}|a{5,6}|bc"
    assert str(parse("a{1,2}|a{4}|a{5,6}").reduce()) == \
        "a{1,2}(a{3,4})?"
    assert str(parse("a{1,2}|a{4}|a{5,6}|bc").reduce()) == \
        "a{1,2}|a{4,6}|bc"
    assert str((parse("a") | parse("a*")).reduce()) == \
        "a*"


def test_bug_28_b():
    # Defect in rxelems.to_fsm()
    assert not parse("(ab*)*").to_fsm().accepts("bb")


def test_derive():
    assert str(parse("a+").derive("a")) == "a*"
    assert str(parse("a+|b+").derive("a")) == "a*"
    assert str(parse("abc|ade").derive("a")) == "bc|de"
    assert str(parse("abc|ade").derive("ab")) == "c"


def test_bug_36_1():
    etc1 = parse(".*").to_fsm()
    etc2 = parse("s.*").to_fsm()
    assert etc1.accepts("s")
    assert etc2.accepts("s")
    assert not etc1.isdisjoint(etc2)
    assert not etc2.isdisjoint(etc1)


def test_bug_36_2():
    etc1 = parse("/etc/.*").to_fsm()
    etc2 = parse("/etc/something.*").to_fsm()
    assert etc1.accepts("/etc/something")
    assert etc2.accepts("/etc/something")
    assert not etc1.isdisjoint(etc2)
    assert not etc2.isdisjoint(etc1)


def test_bug_slow():
    # issue #43
    import time
    m = Fsm(
        alphabet={'R', 'L', 'U', 'D'},
        states={
            0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
            11, 12, 13, 14, 15, 16, 17, 18, 19, 20},
        initial=0,
        finals={20},
        map={
            0: {'D': 1, 'U': 2},
            1: {'L': 3},
            2: {'L': 4},
            3: {'U': 5},
            4: {'D': 6},
            5: {'R': 7},
            6: {'R': 8},
            7: {'U': 9},
            8: {'D': 10},
            9: {'L': 11},
            10: {'L': 12},
            11: {'L': 13},
            12: {'L': 14},
            13: {'D': 15},
            14: {'U': 16},
            15: {'R': 17},
            16: {'R': 18},
            17: {'D': 19},
            18: {'U': 19},
            19: {'L': 20},
            20: {}
        }
    )
    t1 = time.time()
    ll = from_fsm(m)
    t2 = time.time()
    assert (t2 - t1) < 60  # should finish in way under 1s
    assert ll == parse("(DLURULLDRD|ULDRDLLURU)L").reduce()


def test_bug_48_simpler():
    assert str(from_fsm(Fsm(
        alphabet={'d'},
        states={0, 1},
        initial=0,
        finals={1},
        map={
            0: {'d': 1},
        },
    ))) == 'd'


def test_bug_48():
    S5, S26, S45, S63, S80, S97, S113, S127, S140, S152, S163, S175, S182 = \
        range(13)
    char0, char1, char2, char3, char4, char5, char6, char7, char8 = \
        '_', 'a', 'd', 'e', 'g', 'm', 'n', 'o', 'p'

    machine = Fsm(
        alphabet={
            char0, char1, char2, char3, char4,
            char5, char6, char7, char8
        },
        states={
            S5, S26, S45, S63, S80, S97, S113,
            S127, S140, S152, S163, S175, S182
        },
        initial=S5,
        finals={S182},
        map={
            S113: {char0: S127},
            S127: {char7: S140},
            S140: {char6: S152},
            S152: {char0: S163},
            S163: {char5: S175},
            S175: {char8: S182},
            S26: {char1: S45},
            S45: {char5: S63},
            S5: {char2: S26},
            S63: {char1: S80},
            S80: {char4: S97},
            S97: {char3: S113},
        },
    )

    rex = from_fsm(machine)
    assert str(rex) == 'damage_on_mp'


def test_pickle():
    f1 = parse("a{0,4}").to_fsm()
    f2 = parse("a{0,3}").to_fsm()

    assert f2 < f1

    f1_unpickled = pickle.loads(pickle.dumps(f1))
    f2_unpickled = pickle.loads(pickle.dumps(f2))

    assert f2_unpickled < f1_unpickled



# -*- coding: utf-8 -*-

from .bound import Bound
from .multiplier import Multiplier, ZERO, QM, ONE, STAR, PLUS
from .charclass import Charclass
from .rxelems import Conc, Mult, EMPTYSTRING


def test_conc_equality():
    a = Conc(Mult(Charclass("a"), ONE))
    assert a == Conc(Mult(Charclass("a"), ONE))
    assert a != Conc(Mult(Charclass("b"), ONE))
    assert a != Conc(Mult(Charclass("a"), QM))
    assert a != Conc(Mult(
        Charclass("a"),
        Multiplier(Bound(1), Bound(2)))
    )
    assert a != EMPTYSTRING


def test_conc_str():
    assert str(Conc(
        Mult(Charclass("a"), ONE),
        Mult(Charclass("b"), ONE),
        Mult(Charclass("c"), ONE),
        Mult(Charclass("d"), ONE),
        Mult(Charclass("e"), ONE),
        Mult(~Charclass("fg"), STAR),
        Mult(Charclass("h"), Multiplier(Bound(5), Bound(5))),
        Mult(Charclass("abcdefghijklmnopqrstuvwxyz"), PLUS),
    )) == "abcde[^fg]*h{5}[a-z]+"


def test_conc_common():
    a = Mult(Charclass("A"), ONE)
    b = Mult(Charclass("B"), ONE)
    c = Mult(Charclass("C"), ONE)
    y = Mult(Charclass("y"), ONE)
    z = Mult(Charclass("Z"), ONE)
    zstar = Mult(Charclass("Z"), STAR)

    assert Conc(a, a, z, y).common(Conc(b, b, z, y), suffix=True) == Conc(z, y)
    assert Conc(c, z).common(Conc(c, z), suffix=True) == Conc(c, z)
    assert Conc(c, y).common(Conc(c, z), suffix=True) == Conc()
    assert Conc(a, z).common(Conc(b, z), suffix=True) == Conc(z)
    assert Conc(a, zstar).common(Conc(b, z), suffix=True) == Conc()
    assert Conc(a).common(Conc(b), suffix=True) == Conc()


def test_conc_dock():
    a = Mult(Charclass("A"), ONE)
    b = Mult(Charclass("B"), ONE)
    x = Mult(Charclass("X"), ONE)
    x2 = Mult(Charclass("X"), Multiplier(Bound(2), Bound(2)))
    yplus = Mult(Charclass("y"), PLUS)
    z = Mult(Charclass("Z"), ONE)

    assert Conc(a, z).dock(Conc(z)) == Conc(a)
    assert Conc(a, b, x, yplus, z).dock(Conc(x, yplus, z)) == Conc(a, b)
    assert Conc(a, b, x, yplus, z).behead(Conc(a, b, x, yplus)) == Conc(z)
    assert Conc(a).dock(Conc()) == Conc(a)

    try:
        Conc(x2, yplus, z).behead(Conc(x, yplus))
        assert False
    except AssertionError:
        assert False
    except Exception:
        pass


def test_mult_reduction_easy():
    assert Conc(Mult(
        Charclass("a"),
        ZERO
    )).reduce() == Conc()



# -*- coding: utf-8 -*-

from .bound import Bound, INF


def test_bound_str():
    assert str(Bound(2)) == "2"
    assert str(INF) == ""


def test_bound():
    assert min(Bound(0), INF) == Bound(0)
    assert min(Bound(1), INF) == Bound(1)



# -*- coding: utf-8 -*-

from dataclasses import dataclass
from typing import Union, FrozenSet

from .fsm import Fsm, ANYTHING_ELSE


@dataclass(frozen=True)
class Charclass:
    '''
        A `Charclass` is basically a `frozenset` of symbols.
        A `Charclass` with the `negated` flag set is assumed
        to contain every symbol that is in the alphabet of all symbols but not
        explicitly listed inside the frozenset. e.g. [^a]. This is very handy
        if the full alphabet is extremely large, but also requires dedicated
        combination functions.
    '''
    chars: Union[FrozenSet[str], str]
    negated: bool = False

    def __post_init__(self):
        object.__setattr__(self, "chars", frozenset(self.chars))
        # chars should consist only of chars
        if ANYTHING_ELSE in self.chars:
            raise Exception(
                f"Can't put {repr(ANYTHING_ELSE)} in a `Charclass`"
            )

    def __eq__(self, other):
        return isinstance(other, Charclass) \
               and self.chars == other.chars \
               and self.negated == other.negated

    def __hash__(self):
        return hash((self.chars, self.negated))

    # These are the characters carrying special meanings when they appear
    # "outdoors" within a regular expression. To be interpreted literally, they
    # must be escaped with a backslash.
    allSpecial = set("\\[]|().?*+{}")

    # These are the characters carrying special meanings when they appear
    # INSIDE a character class (delimited by square brackets) within a regular
    # expression. To be interpreted literally, they must be escaped with a
    # backslash. Notice how much smaller this class is than the one above; note
    # also that the hyphen and caret do NOT appear above.
    classSpecial = set("\\[]^-")

    # Shorthand codes for use inside `Charclass`es e.g. [abc\d]
    w = frozenset(
        "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz"
    )
    d = frozenset("0123456789")
    s = frozenset("\t\n\v\f\r ")

    shorthand = {
        w: "\\w",
        d: "\\d",
        s: "\\s",
    }

    negated_shorthand = {
        w: "\\W",
        d: "\\D",
        s: "\\S",
    }

    def __str__(self):
        # e.g. \w
        if self in shorthand.keys():
            return shorthand[self]

        # e.g. [^a]
        if self.negated:
            return "[^" + self.escape() + "]"

        # single character, not contained inside square brackets.
        if len(self.chars) == 1:
            # Python lacks the Axiom of Choice
            char = "".join(self.chars)

            # e.g. if char is "\t", return "\\t"
            if char in escapes.keys():
                return escapes[char]

            if char in Charclass.allSpecial:
                return "\\" + char

            # If char is an ASCII control character, don't print it directly,
            # return a hex escape sequence e.g. "\\x00". Note that this
            # includes tab and other characters already handled above
            if 0 <= ord(char) <= 0x1F or ord(char) == 0x7f:
                return "\\x" + "{0:02x}".format(ord(char))

            return char

        # multiple characters (or possibly 0 characters)
        return "[" + self.escape() + "]"

    def escape(self):
        def escapeChar(char):
            if char in Charclass.classSpecial:
                return "\\" + char
            if char in escapes.keys():
                return escapes[char]

            # If char is an ASCII control character, don't print it directly,
            # return a hex escape sequence e.g. "\\x00". Note that this
            # includes tab and other characters already handled above
            if 0 <= ord(char) <= 0x1F or ord(char) == 0x7f:
                return "\\x" + "{0:02x}".format(ord(char))

            return char

        def recordRange():
            # there's no point in putting a range when the whole thing is
            # 3 characters or fewer. "abc" -> "abc" but "abcd" -> "a-d"
            strs = [
                # "ab" or "abc" or "abcd"
                "".join(escapeChar(char) for char in currentRange),
                # "a-b" or "a-c" or "a-d"
                (
                    escapeChar(currentRange[0]) +
                    "-" +
                    escapeChar(currentRange[-1])
                ),
            ]
            return sorted(strs, key=lambda str: len(str))[0]

        output = ""

        # use shorthand for known character ranges
        # note the nested processing order. DO NOT process \d before processing
        # \w. if more character class constants arise which do not nest nicely,
        # a problem will arise because there is no clear ordering to use...

        # look for ranges
        currentRange = ""
        for char in sorted(self.chars, key=ord):

            # range is not empty: new char must fit after previous one
            if len(currentRange) > 0:

                i = ord(char)

                # char doesn't fit old range: restart
                if i != ord(currentRange[-1]) + 1:
                    output += recordRange()
                    currentRange = ""

            currentRange += char

        if len(currentRange) > 0:
            output += recordRange()

        return output

    def to_fsm(self, alphabet=None):
        if alphabet is None:
            alphabet = self.alphabet()

        # 0 is initial, 1 is final

        # If negated, make a singular FSM accepting any other characters
        if self.negated:
            map = {
                0: dict([(symbol, 1) for symbol in alphabet - self.chars]),
            }

        # If normal, make a singular FSM accepting only these characters
        else:
            map = {
                0: dict([(symbol, 1) for symbol in self.chars]),
            }

        return Fsm(
            alphabet=alphabet,
            states={0, 1},
            initial=0,
            finals={1},
            map=map,
        )

    def __repr__(self):
        string = ""
        if self.negated is True:
            string += "~"
        string += "Charclass("
        string += repr("".join(
            str(char) for char in sorted(self.chars, key=str)
        ))
        string += ")"
        return string

    def reduce(self):
        # `Charclass`es cannot be reduced.
        return self

    def alphabet(self):
        return {ANYTHING_ELSE} | self.chars

    def empty(self):
        return len(self.chars) == 0 and not self.negated

    # set operations
    def negate(self):
        '''
            Negate the current `Charclass`. e.g. [ab] becomes [^ab]. Call
            using "charclass2 = ~charclass1"
        '''
        return Charclass(self.chars, negated=not self.negated)

    def __invert__(self):
        return self.negate()

    def reversed(self):
        return self

    def __or__(self, other):
        # ¬A OR ¬B = ¬(A AND B)
        # ¬A OR B = ¬(A - B)
        # A OR ¬B = ¬(B - A)
        # A OR B
        if self.negated:
            if other.negated:
                return ~Charclass(self.chars & other.chars)
            else:
                return ~Charclass(self.chars - other.chars)
        else:
            if other.negated:
                return ~Charclass(other.chars - self.chars)
            else:
                return Charclass(self.chars | other.chars)


# Standard character classes
WORDCHAR = Charclass(
    "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz"
)
DIGIT = Charclass("0123456789")
SPACECHAR = Charclass("\t\n\v\f\r ")

# This `Charclass` expresses "no possibilities at all"
# and can never match anything.
NULLCHARCLASS = Charclass("")

NONWORDCHAR = ~WORDCHAR
NONDIGITCHAR = ~DIGIT
NONSPACECHAR = ~SPACECHAR
DOT = ~NULLCHARCLASS

# Textual representations of standard character classes
shorthand = {
    WORDCHAR: "\\w",
    DIGIT: "\\d",
    SPACECHAR: "\\s",
    NONWORDCHAR: "\\W",
    NONDIGITCHAR: "\\D",
    NONSPACECHAR: "\\S",
    DOT: ".",
}

# Characters which users may escape in a regex instead of inserting them
# literally. In ASCII order:
escapes = {
    "\t": "\\t",  # tab
    "\n": "\\n",  # line feed
    "\v": "\\v",  # vertical tab
    "\f": "\\f",  # form feed
    "\r": "\\r",  # carriage return
}



# -*- coding: utf-8 -*-

# In theory this should just pull in the `parse` function, but it turns out
# that the act of performing any relative import suddenly makes EVERY MODULE
# AVAILABLE IN SCOPE?
# Ignore errors because we are in fact importing these values to re-export them
from .parse import parse  # noqa: F401
from .bound import Bound, INF  # noqa: F401
from .multiplier import Multiplier, QM, STAR, PLUS  # noqa: F401

# WHY ARE THESE MODULE OBJECTS IN SCOPE SUDDENLY? I DIDN'T ASK FOR THEM
# print(bound, fsm, charclass, rxelems)

# This, however, is the *function*, *not* its containing module object?
# print(parse)



# -*- coding: utf-8 -*-

from typing import Tuple

from .bound import Bound, INF
from .charclass import Charclass, shorthand, escapes
from .multiplier import Multiplier, symbolic
from .rxelems import Pattern, Conc, Mult


class NoMatch(Exception):
    '''
        Thrown when parsing fails.
        Almost always caught and almost never fatal
    '''
    pass


def read_until(string: str, i: int, stop_char: str) -> Tuple[int, str]:
    start = i
    while True:
        if i >= len(string):
            raise NoMatch
        if string[i] == stop_char:
            break
        i += 1
    return i + 1, string[start:i]


def static(string, i, static):
    j = i + len(static)
    if string[i:j] == static:
        return j
    raise NoMatch


def select_static(string, i, *statics):
    for st in statics:
        j = i + len(st)
        if string[i:j] == st:
            return j, st
    raise NoMatch


def unescape_hex(string, i):
    '''Turn e.g. "\\x40" into "@". Exactly two hex digits'''
    hex_digits = "0123456789AaBbCcDdEeFf"

    j = static(string, i, "\\x")

    hex1 = string[j]  # e.g. "4"
    if hex1 not in hex_digits:
        raise NoMatch
    j += len(hex1)

    hex2 = string[j]  # e.g. "0"
    if hex2 not in hex_digits:
        raise NoMatch
    j += len(hex2)

    codepoint = int(hex1 + hex2, 16)  # e.g. 64
    char = chr(codepoint)  # "@"
    return char, j


def match_internal_char(string, i):
    # e.g. if we see "\\t", return "\t"
    for key in escapes.keys():
        try:
            return key, static(string, i, escapes[key])
        except NoMatch:
            pass

    # special chars e.g. "\\-" returns "-"
    for char in Charclass.classSpecial:
        try:
            return char, static(string, i, "\\" + char)
        except NoMatch:
            pass

    # hex escape e.g. "\\x40" returns "@"
    try:
        return unescape_hex(string, i)
    except NoMatch:
        pass

    # single non-special character, not contained
    # inside square brackets
    char, j = string[i], i+1
    if char in Charclass.classSpecial:
        raise NoMatch

    return char, j


def match_class_interior_1(string, i):
    # Attempt 1: shorthand e.g. "\w"
    for frozenset, cc_shorthand in Charclass.shorthand.items():
        try:
            return frozenset, False, static(string, i, cc_shorthand)
        except NoMatch:
            pass

    # Attempt 1B: shorthand e.g. "\W"
    for frozenset, cc_shorthand in Charclass.negated_shorthand.items():
        try:
            return frozenset, True, static(string, i, cc_shorthand)
        except NoMatch:
            pass

    # Attempt 2: a range e.g. "d-h"
    try:
        first, j = match_internal_char(string, i)  # `first` is "d"
        k = static(string, j, "-")
        last, k = match_internal_char(string, k)  # `last` is "h"

        firstIndex = ord(first)  # 100
        lastIndex = ord(last)  # 104

        # Be strict here, "d-d" is not allowed
        if firstIndex >= lastIndex:
            raise NoMatch(f"Range '{first}' to '{last}' not allowed")

        chars = set([
            chr(i) for i in range(firstIndex, lastIndex + 1)
        ])
        return chars, False, k
    except NoMatch:
        pass

    # Attempt 3: just a character on its own
    (char, j) = match_internal_char(string, i)
    return set(char), False, j


def match_class_interior(string, i):
    internals = set()
    internals_negated = False
    try:
        while True:
            internal, internal_negated, i = match_class_interior_1(string, i)
            if internal_negated:
                if internals_negated:
                    # E.g. [a1\D\W]
                    internals |= internal
                else:
                    internals_negated = True
                    internals = internal - internals
            else:
                if internals_negated:
                    internals = internals - internal
                else:
                    internals |= internal
    except NoMatch:
        pass
    return internals, internals_negated, i


def match_charclass(string: str, i):
    if i >= len(string):
        raise NoMatch

    # wildcard ".", "\\w", "\\d", etc.
    for key in shorthand.keys():
        try:
            return key, static(string, i, shorthand[key])
        except NoMatch:
            pass

    # "[^dsgsdg]"
    try:
        j = static(string, i, "[^")
        chars, negated, j = match_class_interior(string, j)
        j = static(string, j, "]")
        return Charclass(chars, not negated), j
    except NoMatch:
        pass

    # "[sdfsf]"
    try:
        j = static(string, i, "[")
        chars, negated, j = match_class_interior(string, j)
        j = static(string, j, "]")
        return Charclass(chars, negated), j
    except NoMatch:
        pass

    # e.g. if seeing "\\t", return "\t"
    for key in escapes.keys():
        try:
            return Charclass(key), static(string, i, escapes[key])
        except NoMatch:
            pass

    # e.g. if seeing "\\{", return "{"
    for char in Charclass.allSpecial:
        try:
            return Charclass(char), static(string, i, "\\" + char)
        except NoMatch:
            pass

    # e.g. if seeing "\\x40", return "@"
    try:
        char, j = unescape_hex(string, i)
        return Charclass(char), j
    except NoMatch:
        pass

    # single non-special character, not contained inside square brackets
    char, i = string[i], i+1
    if char in Charclass.allSpecial:
        raise NoMatch

    return Charclass(char), i


def match_multiplicand(string, i):
    # explicitly non-capturing "(?:...)" syntax. No special significance
    try:
        j = static(string, i, "(?")
        j, st = select_static(string, j, ':', 'P<')
        if st == 'P<':
            j, group_name = read_until(string, j, '>')
        pattern, j = match_pattern(string, j)
        j = static(string, j, ")")
        return pattern, j
    except NoMatch:
        pass

    # normal "(...)" syntax
    try:
        j = static(string, i, "(")
        pattern, j = match_pattern(string, j)
        j = static(string, j, ")")
        return pattern, j
    except NoMatch:
        pass

    # Just a `Charclass` on its own
    charclass, j = match_charclass(string, i)
    return charclass, j


def match_any_of(string, i, collection):
    for char in collection:
        try:
            return char, static(string, i, char)
        except NoMatch:
            pass
    raise NoMatch


def match_bound(string: str, i):
    # "0"
    try:
        return Bound(0), static(string, i, "0")
    except NoMatch:
        pass

    # "1", etc.
    try:
        digit, j = match_any_of(string, i, "123456789")
        integer = int(digit)
        try:
            while True:
                digit, j = match_any_of(string, j, "0123456789")
                integer *= 10
                integer += int(digit)
        except NoMatch:
            return Bound(integer), j
    except NoMatch:
        pass

    # "" empty string = infinite bound as in "{4,}"
    return INF, i


def match_multiplier(string: str, i):
    # {2,3} or {2,}
    try:
        j = static(string, i, "{")
        min, j = match_bound(string, j)
        j = static(string, j, ",")
        max, j = match_bound(string, j)
        j = static(string, j, "}")
        return Multiplier(min, max), j
    except NoMatch:
        pass

    # {2}
    try:
        j = static(string, i, "{")
        min, j = match_bound(string, j)
        j = static(string, j, "}")
        return Multiplier(min, min), j
    except NoMatch:
        pass

    # "?"/"*"/"+"/""
    # we do these in reverse order of symbol length, because
    # that forces "" to be done last
    for key in sorted(symbolic, key=lambda key: -len(symbolic[key])):
        try:
            return key, static(string, i, symbolic[key])
        except NoMatch:
            pass

    raise NoMatch


def match_mult(string: str, i):
    multiplicand, j = match_multiplicand(string, i)
    multiplier, j = match_multiplier(string, j)
    return Mult(multiplicand, multiplier), j


def match_conc(string: str, i):
    mults = list()
    try:
        while True:
            m, i = match_mult(string, i)
            mults.append(m)
    except NoMatch:
        pass
    return Conc(*mults), i


def match_pattern(string: str, i):
    concs = list()

    # first one
    c, i = match_conc(string, i)
    concs.append(c)

    # the rest
    while True:
        try:
            i = static(string, i, "|")
            c, i = match_conc(string, i)
            concs.append(c)
        except NoMatch:
            return Pattern(*concs), i


def parse(string: str):
    '''
        Parse a full string and return a `Pattern` object. Fail if
        the whole string wasn't parsed
    '''
    obj, i = match_pattern(string, 0)
    if i != len(string):
        raise Exception(
            f"Could not parse '{string}' beyond index {str(i)}"
        )
    return obj



# -*- coding: utf-8 -*-

'''
    Finite state machine library, intended to be used by `greenery` only
'''

from typing import Optional, Union, Set, Dict
from dataclasses import dataclass

ANYTHING_ELSE = '9bd74361-04f9-4742-9d3a-1d14a6f0044c'


def alphabet_key(symbol):
    '''Ensure `ANYTHING_ELSE` always sorts last'''
    return (symbol is ANYTHING_ELSE, symbol)


class OblivionError(Exception):
    '''
        This exception is thrown while `crawl()`ing an FSM if we transition to
        the oblivion state. For example while crawling two FSMs in parallel we
        may transition to the oblivion state of both FSMs at once. This
        warrants an out-of-band signal which will reduce the complexity of the
        new FSM's map.
    '''
    pass


state_type = Optional[Union[int, str]]


@dataclass(frozen=True)
class Fsm:
    '''
        A Finite State Machine or FSM has an alphabet and a set of states. At
        any given moment, the FSM is in one state. When passed a symbol from
        the alphabet, the FSM jumps to another state (or possibly the same
        state). A map (Python dictionary) indicates where to jump.
        One state is nominated as a starting state. Zero or more states are
        nominated as final states. If, after consuming a string of symbols,
        the FSM is in a final state, then it is said to "accept" the string.
        This class also has some pretty powerful methods which allow FSMs to
        be concatenated, alternated between, multiplied, looped (Kleene star
        closure), intersected, and simplified.
        The majority of these methods are available using operator overloads.
    '''
    initial: state_type
    finals: Set[state_type]
    alphabet: Set[str]
    states: Set[state_type]
    map: Dict[state_type, Dict[str, state_type]]

    def __post_init__(self):
        '''
            `alphabet` is an iterable of symbols the FSM can be fed.
            `states` is the set of states for the FSM
            `initial` is the initial state
            `finals` is the set of accepting states
            `map` may be sparse (i.e. it may omit transitions). In the case of
            omitted transitions, a non-final "oblivion" state is simulated.
        '''

        # Validation. Thanks to immutability, this only needs to be carried out
        # once.
        if self.initial not in self.states:
            raise Exception(
                f"Initial state {repr(self.initial)} "
                f"must be one of {repr(self.states)}"
            )
        if not self.finals.issubset(self.states):
            raise Exception(
                f"Final states {repr(self.finals)} "
                f"must be a subset of {repr(self.states)}"
            )
        for state, symbol in self.map.items():
            for symbol in self.map[state]:
                if not self.map[state][symbol] in self.states:
                    raise Exception(
                        f"Transition for state {repr(state)} "
                        f"and symbol {repr(symbol)} "
                        f"leads to {repr(self.map[state][symbol])}, "
                        "which is not a state"
                    )

        # Initialise the hard way due to immutability.
        object.__setattr__(self, "alphabet", set(self.alphabet))
        object.__setattr__(self, "states", set(self.states))
        object.__setattr__(self, "finals", set(self.finals))

    def accepts(self, input):
        '''
            Test whether the present FSM accepts the supplied string (iterable
            of symbols). Equivalently, consider `self` as a possibly-infinite
            set of strings and test whether `string` is a member of it. This is
            actually mainly used for unit testing purposes. If `ANYTHING_ELSE`
            is in your alphabet, then any symbol not in your alphabet will be
            converted to `ANYTHING_ELSE`.
        '''
        state = self.initial
        for symbol in input:
            if ANYTHING_ELSE in self.alphabet and symbol not in self.alphabet:
                symbol = ANYTHING_ELSE

            # Missing transition = transition to dead state
            if not (state in self.map and symbol in self.map[state]):
                return False

            state = self.map[state][symbol]
        return state in self.finals

    def __contains__(self, string):
        '''
            This lets you use the syntax `"a" in fsm1` to see whether the
            string "a" is in the set of strings accepted by `fsm1`.
        '''
        return self.accepts(string)

    def reduce(self):
        '''
            A result by Brzozowski (1963) shows that a minimal finite state
            machine equivalent to the original can be obtained by reversing the
            original twice.
        '''
        return reversed(reversed(self))

    def __repr__(self):
        args = ", ".join([
            f"alphabet={repr(self.alphabet)}",
            f"states={repr(self.states)}",
            f"initial={repr(self.initial)}",
            f"finals={repr(self.finals)}",
            f"map={repr(self.map)}",
        ])
        return f"Fsm({args})"

    def __str__(self):
        rows = []

        sorted_alphabet = sorted(self.alphabet, key=alphabet_key)

        # top row
        row = ["", "name", "final?"]
        row.extend(
          'ANYTHING_ELSE' if symbol is ANYTHING_ELSE else str(symbol)
          for symbol in sorted_alphabet
        )
        rows.append(row)

        # other rows
        for state in self.states:
            row = []
            if state == self.initial:
                row.append("*")
            else:
                row.append("")
            row.append(str(state))
            if state in self.finals:
                row.append("True")
            else:
                row.append("False")
            for symbol in sorted_alphabet:
                if state in self.map and symbol in self.map[state]:
                    row.append(str(self.map[state][symbol]))
                else:
                    row.append("")
            rows.append(row)

        # column widths
        colwidths = []
        for x in range(len(rows[0])):
            colwidths.append(max(
                len(str(rows[y][x])) for y in range(len(rows))
            ) + 1)

        # apply padding
        for y in range(len(rows)):
            for x in range(len(rows[y])):
                rows[y][x] = rows[y][x].ljust(colwidths[x])

        # horizontal line
        rows.insert(1, ["-" * colwidth for colwidth in colwidths])

        return "".join("".join(row) + "\n" for row in rows)

    def concatenate(*fsms):
        '''
            Concatenate arbitrarily many finite state machines together.
        '''
        alphabet = set().union(*[fsm.alphabet for fsm in fsms])

        def connect_all(i, substate):
            '''
                Take a state in the numbered FSM and return a set containing
                it, plus (if it's final) the first state from the next FSM,
                plus (if that's final) the first state from the next but one
                FSM, plus...
            '''
            result = {(i, substate)}
            while i < len(fsms) - 1 and substate in fsms[i].finals:
                i += 1
                substate = fsms[i].initial
                result.add((i, substate))
            return result

        # Use a superset containing states from all FSMs at once.
        # We start at the start of the first FSM. If this state is final in the
        # first FSM, then we are also at the start of the second FSM. And so
        # on.
        initial = set()
        if len(fsms) > 0:
            initial.update(connect_all(0, fsms[0].initial))
        initial = frozenset(initial)

        def final(state):
            '''If you're in a final state of the final FSM, it's final'''
            for (i, substate) in state:
                if i == len(fsms) - 1 and substate in fsms[i].finals:
                    return True
            return False

        def follow(current, symbol):
            '''
                Follow the collection of states through all FSMs at once,
                jumping to the next FSM if we reach the end of the current one
                TODO: improve all follow() implementations to allow for dead
                metastates?
            '''
            next = set()
            for (i, substate) in current:
                fsm = fsms[i]
                if substate in fsm.map:
                    if symbol in fsm.map[substate]:
                        next.update(connect_all(i, fsm.map[substate][symbol]))
                    elif (
                        ANYTHING_ELSE in fsm.map[substate]
                        and symbol not in fsm.alphabet
                    ):
                        next.update(connect_all(
                            i,
                            fsm.map[substate][ANYTHING_ELSE]
                        ))
            if len(next) == 0:
                raise OblivionError
            return frozenset(next)

        return crawl(alphabet, initial, final, follow).reduce()

    def __add__(self, other):
        '''
            Concatenate two finite state machines together.
            For example, if self accepts "0*" and other accepts "1+(0|1)",
            will return a finite state machine accepting "0*1+(0|1)".
            Accomplished by effectively following non-deterministically.
            Call using "fsm3 = fsm1 + fsm2"
        '''
        return self.concatenate(other)

    def star(self):
        '''
            If the present FSM accepts X, returns an FSM accepting X* (i.e. 0
            or more Xes). This is NOT as simple as naively connecting the final
            states back to the initial state: see (b*ab)* for example.
        '''
        alphabet = self.alphabet

        initial = {self.initial}

        def follow(state, symbol):
            next = set()
            for substate in state:
                if substate in self.map and symbol in self.map[substate]:
                    next.add(self.map[substate][symbol])

                # If one of our substates is final, then we can also consider
                # transitions from the initial state of the original FSM.
                if substate in self.finals \
                   and self.initial in self.map \
                   and symbol in self.map[self.initial]:
                    next.add(self.map[self.initial][symbol])

            if len(next) == 0:
                raise OblivionError

            return frozenset(next)

        def final(state):
            return any(substate in self.finals for substate in state)

        return crawl(alphabet, initial, final, follow) | epsilon(alphabet)

    def times(self, multiplier):
        '''
            Given an FSM and a multiplier, return the multiplied FSM.
        '''
        if multiplier < 0:
            raise Exception(f"Can't multiply an FSM by {repr(multiplier)}")

        alphabet = self.alphabet

        # metastate is a set of iterations+states
        initial = {(self.initial, 0)}

        def final(state):
            '''
                If the initial state is final then multiplying doesn't alter
                that
            '''
            for (substate, iteration) in state:
                if substate == self.initial \
                  and (
                    self.initial in self.finals or
                    iteration == multiplier
                  ):
                    return True
            return False

        def follow(current, symbol):
            next = []
            for (substate, iteration) in current:
                if iteration < multiplier \
                   and substate in self.map \
                   and symbol in self.map[substate]:
                    next.append((self.map[substate][symbol], iteration))
                    # final of self? merge with initial on next iteration
                    if self.map[substate][symbol] in self.finals:
                        next.append((self.initial, iteration + 1))
            if len(next) == 0:
                raise OblivionError
            return frozenset(next)

        return crawl(alphabet, initial, final, follow).reduce()

    def __mul__(self, multiplier):
        '''
            Given an FSM and a multiplier, return the multiplied FSM.
        '''
        return self.times(multiplier)

    def union(*fsms):
        '''
            Treat `fsms` as a collection of arbitrary FSMs and return the union
            FSM. Can be used as `fsm1.union(fsm2, ...)` or
            `fsm.union(fsm1, ...)`. `fsms` may be empty.
        '''
        return parallel(fsms, any)

    def __or__(self, other):
        '''
            Alternation.
            Return a finite state machine which accepts any sequence of symbols
            that is accepted by either self or other. Note that the set of
            strings recognised by the two FSMs undergoes a set union.
            Call using "fsm3 = fsm1 | fsm2"
        '''
        return self.union(other)

    def intersection(*fsms):
        '''
            Intersection.
            Take FSMs and AND them together. That is, return an FSM which
            accepts any sequence of symbols that is accepted by both of the
            original FSMs. Note that the set of strings recognised by the two
            FSMs undergoes a set intersection operation.
            Call using "fsm3 = fsm1 & fsm2"
        '''
        return parallel(fsms, all)

    def __and__(self, other):
        '''
            Treat the FSMs as sets of strings and return the intersection of
            those sets in the form of a new FSM.
        '''
        return self.intersection(other)

    def symmetric_difference(*fsms):
        '''
            Treat `fsms` as a collection of sets of strings and compute the
            symmetric difference of them all. The python set method only allows
            two sets to be operated on at once, but we go the extra mile since
            it's not too hard.
        '''
        return parallel(fsms, lambda accepts: (accepts.count(True) % 2) == 1)

    def __xor__(self, other):
        '''
            Symmetric difference. Returns an FSM which recognises only the
            strings recognised by `self` or `other` but not both.
        '''
        return self.symmetric_difference(other)

    def everythingbut(self):
        '''
            Return a finite state machine which will accept any string NOT
            accepted by self, and will not accept any string accepted by self.
            This is more complicated if there are missing transitions, because
            the missing "dead" state must now be reified.
        '''
        alphabet = self.alphabet

        initial = {0: self.initial}

        def follow(current, symbol):
            next = {}
            if 0 in current \
               and current[0] in self.map \
               and symbol in self.map[current[0]]:
                next[0] = self.map[current[0]][symbol]
            return next

        # state is final unless the original was
        def final(state):
            return not (0 in state and state[0] in self.finals)

        return crawl(alphabet, initial, final, follow).reduce()

    def reversed(self):
        '''
            Return a new FSM such that for every string that self accepts (e.g.
            "beer", the new FSM accepts the reversed string ("reeb").
        '''
        alphabet = self.alphabet

        # Start from a composite "state-set" consisting of all final states.
        # If there are no final states, this set is empty and we'll find that
        # no other states get generated.
        initial = frozenset(self.finals)

        # Find every possible way to reach the current state-set
        # using this symbol.
        def follow(current, symbol):
            next = frozenset([
                prev
                for prev in self.map
                for state in current
                if symbol in self.map[prev] and self.map[prev][symbol] == state
            ])
            if len(next) == 0:
                raise OblivionError
            return next

        # A state-set is final if the initial state is in it.
        def final(state):
            return self.initial in state

        # Man, crawl() is the best!
        return crawl(alphabet, initial, final, follow)
        # Do not reduce() the result, since reduce() calls us in turn

    def __reversed__(self):
        '''
            Return a new FSM such that for every string that self accepts (e.g.
            "beer", the new FSM accepts the reversed string ("reeb").
        '''
        return self.reversed()

    def islive(self, state):
        '''A state is "live" if a final state can be reached from it.'''
        reachable = [state]
        i = 0
        while i < len(reachable):
            current = reachable[i]
            if current in self.finals:
                return True
            if current in self.map:
                for symbol in self.map[current]:
                    next = self.map[current][symbol]
                    if next not in reachable:
                        reachable.append(next)
            i += 1
        return False

    def empty(self):
        '''
            An FSM is empty if it recognises no strings. An FSM may be
            arbitrarily complicated and have arbitrarily many final states
            while still recognising no strings because those final states may
            all be inaccessible from the initial state. Equally, an FSM may be
            non-empty despite having an empty alphabet if the initial state is
            final.
        '''
        return not self.islive(self.initial)

    def strings(self):
        '''
            Generate strings (lists of symbols) that this FSM accepts. Since
            there may be infinitely many of these we use a generator instead of
            constructing a static list. Strings will be sorted in order of
            length and then lexically. This procedure uses arbitrary amounts of
            memory but is very fast. There may be more efficient ways to do
            this, that I haven't investigated yet. You can use this in list
            comprehensions.
        '''

        # Many FSMs have "dead states". Once you reach a dead state, you can no
        # longer reach a final state. Since many strings may end up here, it's
        # advantageous to constrain our search to live states only.
        livestates = set(state for state in self.states if self.islive(state))

        # We store a list of tuples. Each tuple consists of an input string and
        # the state that this input string leads to. This means we don't have
        # to run the state machine from the very beginning every time we want
        # to check a new string.
        strings = []

        # Initial entry (or possibly not, in which case this is a short one)
        cstate = self.initial
        cstring = []
        if cstate in livestates:
            if cstate in self.finals:
                yield cstring
            strings.append((cstring, cstate))

        # Fixed point calculation
        i = 0
        while i < len(strings):
            (cstring, cstate) = strings[i]
            if cstate in self.map:
                for symbol in sorted(self.map[cstate], key=alphabet_key):
                    nstate = self.map[cstate][symbol]
                    nstring = cstring + [symbol]
                    if nstate in livestates:
                        if nstate in self.finals:
                            yield nstring
                        strings.append((nstring, nstate))
            i += 1

    def __iter__(self):
        '''
            This allows you to do `for string in fsm1` as a list comprehension!
        '''
        return self.strings()

    def equivalent(self, other):
        '''
            Two FSMs are considered equivalent if they recognise the same
            strings. Or, to put it another way, if their symmetric difference
            recognises no strings.
        '''
        return (self ^ other).empty()

    def __eq__(self, other):
        '''
            You can use `fsm1 == fsm2` to determine whether two FSMs recognise
            the same strings.
        '''
        return self.equivalent(other)

    def different(self, other):
        '''
            Two FSMs are considered different if they have a non-empty
            symmetric difference.
        '''
        return not (self ^ other).empty()

    def __ne__(self, other):
        '''
            Use `fsm1 != fsm2` to determine whether two FSMs recognise
            different strings.
        '''
        return self.different(other)

    def difference(*fsms):
        '''
            Difference. Returns an FSM which recognises only the strings
            recognised by the first FSM in the list, but none of the others.
        '''
        return parallel(
            fsms,
            lambda accepts: accepts[0] and not any(accepts[1:])
        )

    def __sub__(self, other):
        return self.difference(other)

    def cardinality(self):
        '''
            Consider the FSM as a set of strings and return the cardinality of
            that set, or raise an OverflowError if there are infinitely many
        '''
        num_strings = {}

        def get_num_strings(state):
            # Many FSMs have at least one oblivion state
            if self.islive(state):
                if state in num_strings:
                    if num_strings[state] is None:  # "computing..."
                        # Recursion! There are infinitely many strings
                        # recognised
                        raise OverflowError(state)
                    return num_strings[state]
                num_strings[state] = None  # i.e. "computing..."

                n = 0
                if state in self.finals:
                    n += 1
                if state in self.map:
                    for symbol in self.map[state]:
                        n += get_num_strings(self.map[state][symbol])
                num_strings[state] = n

            else:
                # Dead state
                num_strings[state] = 0

            return num_strings[state]

        return get_num_strings(self.initial)

    def __len__(self):
        '''
            Consider the FSM as a set of strings and return the cardinality of
            that set, or raise an OverflowError if there are infinitely many
        '''
        return self.cardinality()

    def isdisjoint(self, other):
        '''
            Treat `self` and `other` as sets of strings and see if they are
            disjoint
        '''
        return (self & other).empty()

    def issubset(self, other):
        '''
            Treat `self` and `other` as sets of strings and see if `self` is a
            subset of `other`... `self` recognises no strings which `other`
            doesn't.
        '''
        return (self - other).empty()

    def __le__(self, other):
        '''
            Treat `self` and `other` as sets of strings and see if `self` is a
            subset of `other`... `self` recognises no strings which `other`
            doesn't.
        '''
        return self.issubset(other)

    def ispropersubset(self, other):
        '''
            Treat `self` and `other` as sets of strings and see if `self` is a
            proper subset of `other`.
        '''
        return self <= other and self != other

    def __lt__(self, other):
        '''
            Treat `self` and `other` as sets of strings and see if `self` is a
            strict subset of `other`.
        '''
        return self.ispropersubset(other)

    def issuperset(self, other):
        '''
            Treat `self` and `other` as sets of strings and see if `self` is a
            superset of `other`.
        '''
        return (other - self).empty()

    def __ge__(self, other):
        '''
            Treat `self` and `other` as sets of strings and see if `self` is a
            superset of `other`.
        '''
        return self.issuperset(other)

    def ispropersuperset(self, other):
        '''
            Treat `self` and `other` as sets of strings and see if `self` is a
            proper superset of `other`.
        '''
        return self >= other and self != other

    def __gt__(self, other):
        '''
            Treat `self` and `other` as sets of strings and see if `self` is a
            strict superset of `other`.
        '''
        return self.ispropersuperset(other)

    def copy(self):
        '''
            For completeness only, since `set.copy()` also exists. FSM objects
            are immutable, so I can see only very odd reasons to need this.
        '''
        return Fsm(
            alphabet=self.alphabet,
            states=self.states,
            initial=self.initial,
            finals=self.finals,
            map=self.map,
        )

    def derive(self, input):
        '''
            Compute the Brzozowski derivative of this FSM with respect to the
            input string of symbols.
            <https://en.wikipedia.org/wiki/Brzozowski_derivative>
            If any of the symbols are not members of the alphabet, that's a
            `KeyError`. If you fall into oblivion, then the derivative is an
            FSM accepting no strings.
        '''
        try:
            # Consume the input string.
            state = self.initial
            for symbol in input:
                if symbol not in self.alphabet:
                    if ANYTHING_ELSE not in self.alphabet:
                        raise KeyError(symbol)
                    symbol = ANYTHING_ELSE

                # Missing transition = transition to dead state
                if not (state in self.map and symbol in self.map[state]):
                    raise OblivionError

                state = self.map[state][symbol]

            # OK so now we have consumed that string, use the new location as
            # the starting point.
            return Fsm(
                alphabet=self.alphabet,
                states=self.states,
                initial=state,
                finals=self.finals,
                map=self.map,
            )

        except OblivionError:
            # Fell out of the FSM. The derivative of this FSM is the empty FSM.
            return null(self.alphabet)


def null(alphabet):
    '''
        An FSM accepting nothing (not even the empty string). This is
        demonstrates that this is possible, and is also extremely useful
        in some situations
    '''
    return Fsm(
        alphabet=alphabet,
        states={0},
        initial=0,
        finals=set(),
        map={
            0: dict([(symbol, 0) for symbol in alphabet]),
        },
    )


def epsilon(alphabet):
    '''
        Return an FSM matching an empty string, "", only.
        This is very useful in many situations
    '''
    return Fsm(
        alphabet=alphabet,
        states={0},
        initial=0,
        finals={0},
        map={},
    )


def parallel(fsms, test):
    '''
        Crawl several FSMs in parallel, mapping the states of a larger
        meta-FSM. To determine whether a state in the larger FSM is final, pass
        all of the finality statuses (e.g. [True, False, False] to `test`.
    '''
    alphabet = set().union(*[fsm.alphabet for fsm in fsms])

    initial = dict([(i, fsm.initial) for (i, fsm) in enumerate(fsms)])

    # dedicated function accepts a "superset" and returns the next "superset"
    # obtained by following this transition in the new FSM
    def follow(current, symbol):
        next = {}
        for i in range(len(fsms)):
            if symbol not in fsms[i].alphabet \
               and ANYTHING_ELSE in fsms[i].alphabet:
                actual_symbol = ANYTHING_ELSE
            else:
                actual_symbol = symbol
            if i in current \
               and current[i] in fsms[i].map \
               and actual_symbol in fsms[i].map[current[i]]:
                next[i] = fsms[i].map[current[i]][actual_symbol]
        if len(next.keys()) == 0:
            raise OblivionError
        return next

    # Determine the "is final?" condition of each substate, then pass it to the
    # test to determine finality of the overall FSM.
    def final(state):
        accepts = [
            i in state and state[i] in fsm.finals
            for (i, fsm) in enumerate(fsms)
        ]
        return test(accepts)

    return crawl(alphabet, initial, final, follow).reduce()


def crawl(alphabet, initial, final, follow):
    '''
        Given the above conditions and instructions, crawl a new unknown FSM,
        mapping its states, final states and transitions. Return the new FSM.
        This is a pretty powerful procedure which could potentially go on
        forever if you supply an evil version of follow().
    '''

    states = [initial]
    finals = set()
    map = {}

    # iterate over a growing list
    i = 0
    while i < len(states):
        state = states[i]

        # add to finals
        if final(state):
            finals.add(i)

        # compute map for this state
        map[i] = {}
        for symbol in sorted(alphabet, key=alphabet_key):
            try:
                next = follow(state, symbol)

                try:
                    j = states.index(next)
                except ValueError:
                    j = len(states)
                    states.append(next)

            except OblivionError:
                # Reached an oblivion state. Don't list it.
                continue

            map[i][symbol] = j

        i += 1

    return Fsm(
        alphabet=alphabet,
        states=set(range(len(states))),
        initial=0,
        finals=finals,
        map=map,
    )



# -*- coding: utf-8 -*-

from .charclass import Charclass
from .multiplier import ONE, ZERO
from .rxelems import Pattern, Conc, Mult
from .parse import parse


def test_pattern_equality():
    assert Pattern(
        Conc(Mult(Charclass("a"), ONE)),
        Conc(Mult(Charclass("b"), ONE)),
    ) == Pattern(
        Conc(Mult(Charclass("b"), ONE)),
        Conc(Mult(Charclass("a"), ONE)),
    )
    assert Pattern(
        Conc(Mult(Charclass("a"), ONE)),
        Conc(Mult(Charclass("a"), ONE)),
    ) == Pattern(
        Conc(Mult(Charclass("a"), ONE)),
    )


def test_pattern_str():
    assert str(Pattern(
        Conc(Mult(Charclass("a"), ONE)),
        Conc(Mult(Charclass("b"), ONE)),
    )) == "a|b"
    assert str(Pattern(
        Conc(Mult(Charclass("a"), ONE)),
        Conc(Mult(Charclass("a"), ONE)),
    )) == "a"
    assert str(Pattern(
        Conc(
            Mult(Charclass("a"), ONE),
            Mult(Charclass("b"), ONE),
            Mult(Charclass("c"), ONE),
        ),
        Conc(
            Mult(Charclass("d"), ONE),
            Mult(Charclass("e"), ONE),
            Mult(Charclass("f"), ONE),
            Mult(
                Pattern(
                    Conc(
                        Mult(Charclass("g"), ONE),
                        Mult(Charclass("h"), ONE),
                        Mult(Charclass("i"), ONE),
                    ),
                    Conc(
                        Mult(Charclass("j"), ONE),
                        Mult(Charclass("k"), ONE),
                        Mult(Charclass("l"), ONE),
                    ),
                ),
                ONE,
            ),
        ),
    )) == "abc|def(ghi|jkl)"


def test_empty():
    assert Pattern().empty()


def test_mult_reduction_easy():
    assert Pattern(Conc()).reduce() == Pattern(Conc())
    assert Pattern(
        Conc(
            Mult(
                Charclass("a"),
                ZERO,
            )
        )
    ).reduce() == Pattern(
        Conc()
    )
    assert str(
        Pattern(
            Conc(
                Mult(
                    Charclass("a"),
                    ZERO,
                )
            )
        ).reduce()
    ) == ""


def test_empty_pattern_reduction():
    assert str(Pattern().reduce()) == "[]"


def test_empty_conc_suppression():
    assert str(Pattern(
        Conc(
            # this `Mult` can never actually match anything
            Mult(Pattern(), ONE),
            Mult(Charclass("0"), ONE),
            Mult(Charclass("0123456789"), ONE),
        )  # so neither can this `Conc`
    ).reduce()) == "[]"


def test_pattern_dock():
    a = Mult(Charclass("a"), ONE)
    c = Mult(Charclass("c"), ONE)
    f = Mult(Charclass("f"), ONE)

    assert parse("a|bc").dock(Conc()) == parse("a|bc")
    assert parse("aa|bca").dock(Conc(a)) == parse("a|bc")
    assert parse("xyza|abca|a").dock(Conc(a)) == parse("xyz|abc|")
    assert parse("f{2,3}c|fc").dock(Conc(f, c)) == parse("f{1,2}|")
    assert parse("aa").dock(Conc(a, a)) == parse("")


def test_pattern_beheading():
    a = Mult(Charclass("a"), ONE)
    c = Mult(Charclass("c"), ONE)
    f = Mult(Charclass("f"), ONE)
    z = Mult(Charclass("Z"), ONE)

    assert parse("aa").behead(Conc(a)) == parse("a")
    assert parse("abc|aa").behead(Conc(a)) == parse("a|bc")
    assert parse("cf{1,2}|cf").behead(Conc(c)) == parse("f{1,2}|f")
    assert parse("aa|aa").behead(Conc(a, a)) == parse("")
    assert parse("abc|aa").behead(Conc(a)) == parse("a|bc")
    assert parse("a|bc").behead(Conc()) == parse("a|bc")
    assert parse("cf{1,2}|cf").behead(Conc(c, f)) == parse("f?|")
    assert parse("ZA|ZB|ZC").behead(Conc(z)) == parse("A|B|C")
    assert parse("Z+A|ZB|ZZC").behead(Conc(z)) == parse("Z*A|B|ZC")
    assert parse("a{2}b|a+c").behead(Conc(a)) == parse("ab|a*c")



# -*- coding: utf-8 -*-

from .bound import Bound, INF
from .multiplier import Multiplier, ZERO, ONE, QM, STAR, PLUS


def test_multiplier_str():
    assert str(Multiplier(Bound(2), INF)) == "{2,}"


def test_bound_qm():
    assert QM.mandatory == Bound(0)
    assert QM.optional == Bound(1)


def test_multiplier_common():
    assert ZERO.common(ZERO) == ZERO
    assert ZERO.common(QM) == ZERO
    assert ZERO.common(ONE) == ZERO
    assert ZERO.common(STAR) == ZERO
    assert ZERO.common(PLUS) == ZERO
    assert QM.common(ZERO) == ZERO
    assert QM.common(QM) == QM
    assert QM.common(ONE) == ZERO
    assert QM.common(STAR) == QM
    assert QM.common(PLUS) == QM
    assert ONE.common(ZERO) == ZERO
    assert ONE.common(QM) == ZERO
    assert ONE.common(ONE) == ONE
    assert ONE.common(STAR) == ZERO
    assert ONE.common(PLUS) == ONE
    assert STAR.common(ZERO) == ZERO
    assert STAR.common(QM) == QM
    assert STAR.common(ONE) == ZERO
    assert STAR.common(STAR) == STAR
    assert STAR.common(PLUS) == STAR
    assert PLUS.common(ZERO) == ZERO
    assert PLUS.common(QM) == QM
    assert PLUS.common(ONE) == ONE
    assert PLUS.common(STAR) == STAR
    assert PLUS.common(PLUS) == PLUS


def test_multiplier_subtraction():
    # a{3,4}, a{2,5} -> a{2,3} (with a{1,1}, a{0,2} left over)
    assert Multiplier(Bound(3), Bound(4)) \
        .common(Multiplier(Bound(2), Bound(5))) \
        == Multiplier(Bound(2), Bound(3))
    assert Multiplier(Bound(3), Bound(4)) - Multiplier(Bound(2), Bound(3)) \
        == ONE
    assert Multiplier(Bound(2), Bound(5)) - Multiplier(Bound(2), Bound(3)) \
        == Multiplier(Bound(0), Bound(2))

    # a{2,}, a{1,5} -> a{1,5} (with a{1,}, a{0,0} left over)
    assert Multiplier(Bound(2), INF).common(Multiplier(Bound(1), Bound(5))) \
        == Multiplier(Bound(1), Bound(5))
    assert Multiplier(Bound(2), INF) - Multiplier(Bound(1), Bound(5)) \
        == PLUS
    assert Multiplier(Bound(1), Bound(5)) - Multiplier(Bound(1), Bound(5)) \
        == ZERO

    # a{3,}, a{2,} -> a{2,} (with a, epsilon left over)
    assert Multiplier(Bound(3), INF).common(Multiplier(Bound(2), INF)) \
        == Multiplier(Bound(2), INF)
    assert Multiplier(Bound(3), INF) - Multiplier(Bound(2), INF) == ONE
    assert Multiplier(Bound(2), INF) - Multiplier(Bound(2), INF) == ZERO

    # a{3,}, a{3,} -> a{3,} (with ZERO, ZERO left over)
    assert Multiplier(Bound(3), INF).common(Multiplier(Bound(3), INF)) \
        == Multiplier(Bound(3), INF)
    assert Multiplier(Bound(3), INF) - Multiplier(Bound(3), INF) == ZERO


def test_multiplier_union():
    assert ZERO | ZERO == ZERO
    assert ZERO | QM == QM
    assert ZERO | ONE == QM
    assert ZERO | STAR == STAR
    assert ZERO | PLUS == STAR
    assert QM | ZERO == QM
    assert QM | QM == QM
    assert QM | ONE == QM
    assert QM | STAR == STAR
    assert QM | PLUS == STAR
    assert ONE | ZERO == QM
    assert ONE | QM == QM
    assert ONE | ONE == ONE
    assert ONE | STAR == STAR
    assert ONE | PLUS == PLUS
    assert STAR | ZERO == STAR
    assert STAR | QM == STAR
    assert STAR | ONE == STAR
    assert STAR | STAR == STAR
    assert STAR | PLUS == STAR
    assert PLUS | ZERO == STAR
    assert PLUS | QM == STAR
    assert PLUS | ONE == PLUS
    assert PLUS | STAR == STAR
    assert PLUS | PLUS == PLUS
    assert not ZERO.canunion(Multiplier(Bound(2), INF))
    assert not ONE.canunion(Multiplier(Bound(3), Bound(4)))
    assert not Multiplier(
        Bound(8),
        INF
    ).canunion(
        Multiplier(
            Bound(3),
            Bound(4)
        )
    )
    try:
        ZERO | Multiplier(Bound(7), Bound(8))
        assert False
    except AssertionError:
        assert False
    except Exception:
        pass



from setuptools import setup

setup(
    name="greenery",
    version="4.0.0",
    tests_require=["pytest"],
    packages=["greenery"],
    package_dir={
        "greenery": "greenery"
    },
    author="qntm",
    author_email="qntm <qntm@users.noreply.github.com>",
    description="Greenery allows manipulation of regular expressions",
    license="MIT License",
    keywords=" ".join([
        "re", "regex", "regexp", "regular", "expression", "deterministic",
        "finite", "state", "machine", "automaton", "fsm", "dfsm", "fsa",
        "dfsa", "greenery",
    ]),
    url="https://github.com/qntm/greenery",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
    ],
)



# -*- coding: utf-8 -*-

from .fsm import ANYTHING_ELSE
from .charclass import Charclass, DIGIT
from .bound import Bound, INF
from .multiplier import Multiplier, ONE, QM, STAR, PLUS
from .rxelems import Mult


def test_mult_equality():
    a = Mult(Charclass("a"), ONE)
    assert a == a
    assert a != Mult(Charclass("b"), ONE)
    assert a != Mult(Charclass("a"), QM)
    assert a != Mult(
        Charclass("a"),
        Multiplier(Bound(1), Bound(2))
    )


def test_mult_str():
    a = Charclass("a")
    assert str(Mult(a, ONE)) == "a"
    assert str(Mult(a, Multiplier(Bound(2), Bound(2)))) == "a{2}"
    assert str(Mult(a, Multiplier(Bound(3), Bound(3)))) == "a{3}"
    assert str(Mult(a, Multiplier(Bound(4), Bound(4)))) == "a{4}"
    assert str(Mult(a, Multiplier(Bound(5), Bound(5)))) == "a{5}"
    assert str(Mult(a, QM)) == "a?"
    assert str(Mult(a, STAR)) == "a*"
    assert str(Mult(a, PLUS)) == "a+"
    assert str(Mult(a, Multiplier(Bound(2), Bound(5)))) == "a{2,5}"
    assert str(Mult(a, Multiplier(Bound(2), INF))) == "a{2,}"

    assert str(Mult(DIGIT, ONE)) == "\\d"
    assert str(Mult(DIGIT, Multiplier(Bound(2), Bound(2)))) == "\\d{2}"
    assert str(Mult(DIGIT, Multiplier(Bound(3), Bound(3)))) == "\\d{3}"


def test_odd_bug():
    # Odd bug with ([bc]*c)?[ab]*
    int5A = Mult(
        Charclass("bc"),
        STAR,
    ).to_fsm({"a", "b", "c", ANYTHING_ELSE})
    assert int5A.accepts([])
    assert int5A.accepts("")

    int5B = Mult(
        Charclass("c"),
        ONE,
    ).to_fsm({"a", "b", "c", ANYTHING_ELSE})
    assert int5B.accepts("c")
    assert int5B.accepts(["c"])

    int5C = int5A + int5B
    assert int5C.accepts("c")
    assert int5C.accepts(["c"])


def test_mult_common():
    a = Charclass("a")
    assert Mult(a, Multiplier(Bound(3), Bound(4))) \
        .common(Mult(a, Multiplier(Bound(2), Bound(5)))) == \
        Mult(a, Multiplier(Bound(2), Bound(3)))
    assert Mult(a, Multiplier(Bound(2), INF)) \
        .common(Mult(a, Multiplier(Bound(1), Bound(5)))) == \
        Mult(a, Multiplier(Bound(1), Bound(5)))
    assert Mult(a, Multiplier(Bound(3), INF)) \
        .common(Mult(a, Multiplier(Bound(2), INF))) == \
        Mult(a, Multiplier(Bound(2), INF))


def test_mult_dock():
    a = Charclass("a")
    assert Mult(a, Multiplier(Bound(4), Bound(5))) \
        .dock(Mult(a, Multiplier(Bound(3), Bound(3)))) == \
        Mult(a, Multiplier(Bound(1), Bound(2)))



# -*- coding: utf-8 -*-

from dataclasses import dataclass, field

from .bound import Bound, INF


@dataclass(frozen=True)
class Multiplier:
    '''
        A min and a max. The vast majority of characters in regular expressions
        occur without a specific multiplier, which is implicitly equivalent to
        a min of 1 and a max of 1, but many more have explicit multipliers like
        "*" (min = 0, max = inf) and so on.

        Although it seems odd and can lead to some confusing edge cases, we do
        also permit a max of 0 (iff min is 0 too). This allows the multiplier
        `ZERO` to exist, which actually are quite useful in their own special
        way.
    '''
    min: Bound
    max: Bound
    mandatory: Bound = field(init=False)
    optional: Bound = field(init=False)

    def __post_init__(self):
        if self.min == INF:
            raise Exception(
                f"Minimum bound of a multiplier can't be {repr(INF)}"
            )
        if self.min > self.max:
            raise Exception(
                "Invalid multiplier bounds: "
                f"{repr(self.min)} and {repr(self.max)}"
            )

        # More useful than "min" and "max" in many situations
        # are "mandatory" and "optional".
        object.__setattr__(self, "mandatory", self.min)
        object.__setattr__(self, "optional", self.max - self.min)

    def __eq__(self, other):
        return self.min == other.min and self.max == other.max

    def __hash__(self):
        return hash((self.min, self.max))

    def __repr__(self):
        return f"Multiplier({repr(self.min)}, {repr(self.max)})"

    def __str__(self):
        if self.max == Bound(0):
            raise Exception(
                f"Can't serialise a multiplier with bound {repr(self.max)}"
            )
        if self in symbolic.keys():
            return symbolic[self]
        if self.min == self.max:
            return "{" + str(self.min) + "}"
        return "{" + str(self.min) + "," + str(self.max) + "}"

    def canmultiplyby(self, other):
        '''
            Multiplication is not well-defined for all pairs of multipliers
            because the resulting possibilities do not necessarily form a
            continuous range.

            For example:
                {0,x} * {0,y} = {0,x*y}
                {2} * {3} = {6}
                {2} * {1,2} = ERROR

            The proof isn't simple but suffice it to say that {p,p+q} * {r,r+s}
            is equal to {pr, (p+q)(r+s)} only if s=0 or qr+1 >= p. If not, then
            at least one gap appears in the range. The first inaccessible
            number is (p+q)r+1. And no, multiplication is not commutative
        '''
        return other.optional == Bound(0) or \
            self.optional * other.mandatory + Bound(1) >= self.mandatory

    def __mul__(self, other):
        '''Multiply this multiplier by another'''
        if not self.canmultiplyby(other):
            raise Exception(
                f"Can't multiply {repr(self)} by {repr(other)}"
            )
        return Multiplier(self.min * other.min, self.max * other.max)

    def __add__(self, other):
        '''Add two multipliers together'''
        return Multiplier(self.min + other.min, self.max + other.max)

    def __sub__(self, other):
        '''
            Subtract another multiplier from this one.
            Caution: multipliers are not totally ordered.
            This operation is not meaningful for all pairs of multipliers.
        '''
        mandatory = self.mandatory - other.mandatory
        optional = self.optional - other.optional
        return Multiplier(mandatory, mandatory + optional)

    def canintersect(self, other):
        '''
            Intersection is not well-defined for all pairs of multipliers.
            For example:
                {2,3} & {3,4} = {3}
                {2,} & {1,7} = {2,7}
                {2} & {5} = ERROR
        '''
        return not (self.max < other.min or other.max < self.min)

    def __and__(self, other):
        '''
            Find the intersection of two multipliers: that is, a third
            multiplier expressing the range covered by both of the originals.
            This is not defined for all multipliers since they may not overlap.
        '''
        if not self.canintersect(other):
            raise Exception(
                f"Can't compute intersection of {repr(self)} and {repr(other)}"
            )
        a = max(self.min, other.min)
        b = min(self.max, other.max)
        return Multiplier(a, b)

    def canunion(self, other):
        '''
            Union is not defined for all pairs of multipliers.
            E.g. {0,1} | {3,4} -> nope
        '''
        return not (
            self.max + Bound(1) < other.min or
            other.max + Bound(1) < self.min
        )

    def __or__(self, other):
        '''
            Find the union of two multipliers: that is, a third multiplier
            expressing the range covered by either of the originals. This is
            not defined for all multipliers since they may not intersect.
        '''
        if not self.canunion(other):
            raise Exception(
                f"Can't compute the union of {repr(self)} and {repr(other)}"
            )
        a = min(self.min, other.min)
        b = max(self.max, other.max)
        return Multiplier(a, b)

    def common(self, other):
        '''
            Find the shared part of two multipliers. This is the largest
            multiplier which can be safely subtracted from both the originals.
            This may return the `ZERO` multiplier.
        '''
        mandatory = min(self.mandatory, other.mandatory)
        optional = min(self.optional, other.optional)
        return Multiplier(mandatory, mandatory + optional)

    def copy(self):
        return Multiplier(self.min.copy(), self.max.copy())


# Preset multipliers. These get used ALL THE TIME in unit tests
ZERO = Multiplier(Bound(0), Bound(0))  # has some occasional uses
QM = Multiplier(Bound(0), Bound(1))
ONE = Multiplier(Bound(1), Bound(1))
STAR = Multiplier(Bound(0), INF)
PLUS = Multiplier(Bound(1), INF)

# Symbol lookup table for preset multipliers.
symbolic = {
    QM: "?",
    ONE: "",
    STAR: "*",
    PLUS: "+",
}
